"""
This is a script for the frontend of the AAclust class, a clustering wrapper object to obtain redundancy-reduced
scale subsets.
"""
from typing import Optional, Dict, List, Tuple, Type
import numpy as np
from sklearn.cluster import KMeans
from sklearn.base import ClusterMixin
from sklearn.exceptions import ConvergenceWarning
import warnings
import pandas as pd

from aaanalysis.template_classes import Wrapper
import aaanalysis.utils as ut

from ._backend.check_aaclust import check_metric
from ._backend.aaclust.aaclust_fit import estimate_lower_bound_n_clusters, optimize_n_clusters, merge_clusters
from ._backend.aaclust.aaclust_eval import evaluate_clustering
from ._backend.aaclust.aaclust_methods import (compute_centers, compute_medoids, name_clusters,
                                               compute_correlation)


# I Helper Functions
# Check parameter matching functions
def check_match_X_names(X=None, names=None, accept_none=True):
    """Verify that the number of samples in 'X' matches the length of 'names'."""
    if accept_none and names is None:
        return
    n_samples, n_features = X.shape
    if n_samples != len(names):
        raise ValueError(f"n_samples does not match for 'X' ({len(X)}) and 'names' ({len(names)}).")


def check_match_X_n_clusters(X=None, n_clusters=None, accept_none=True):
    """Ensure the number of samples and unique samples in 'X' are greater than or equal to 'n_clusters'."""
    if accept_none and n_clusters is None:
        return
    n_samples, n_features = X.shape
    n_unique_samples = len(set(map(tuple, X)))
    if n_samples < n_clusters:
        raise ValueError(f"n_samples={n_samples} (in 'X') should be >= 'n_clusters' ({n_clusters})")
    if n_unique_samples < n_clusters:
        raise ValueError(f"'n_clusters' ({n_clusters}) should be >= n_unique_samples={n_unique_samples} (in 'X').")


def check_X_X_ref(X=None, X_ref=None):
    """Check that the number of features in 'X' matches the number of features in 'X_ref'."""
    n_samples, n_features = X.shape
    n_samples_ref, n_features_ref = X_ref.shape
    if n_features != n_features_ref:
        raise ValueError(f"n_features does not match for 'X' ({n_features}) and 'X_ref' ({n_features_ref}).")


def check_labels_cor(labels=None, labels_name="labels"):
    """Validate that 'labels' contains only integer values and is not None."""
    if labels is None:
        raise ValueError(f"'{labels_name}' should not be None.")
    # Convert labels to a numpy array if it's not already
    labels = np.asarray(labels)
    unique_labels = set(labels)
    wrong_types = [l for l in unique_labels if not np.issubdtype(type(l), np.integer)]
    if wrong_types:
        raise ValueError(f"Labels in '{labels_name}' should be type int, but contain: {set(map(type, wrong_types))}")
    return labels


# Post check functions
def post_check_n_clusters(n_clusters_actual=None, n_clusters=None):
    """Check if n_clusters set properly"""
    if n_clusters is not None and n_clusters_actual < n_clusters:
        warnings.warn(f"'n_clusters' was reduced from {n_clusters} to {n_clusters_actual} "
                      f"during AAclust algorithm.", ConvergenceWarning)


# II Main Functions
class AAclust(Wrapper):
    """
    Amino Acid clustering (**AAclust**) class: A k-optimized clustering wrapper for selecting redundancy-reduced sets
    of numerical scales [Breimann24a]_.

    AAclust uses clustering models that require a pre-defined number of clusters (k, set by ``n_clusters``),
    such as k-means or other `scikit-learn clustering models <https://scikit-learn.org/stable/modules/clustering.html>`_.
    It optimizes the value of k by utilizing Pearson correlation and then selects a representative sample ('medoid')
    for each cluster closest to the center, resulting in a redundancy-reduced sample set.

    Attributes
    ----------
    model : object
        The fitted clustering model object after calling the ``fit`` method.
    n_clusters : int
        Number of clusters obtained by AAclust.
    labels_ : array-like, shape (n_samples)
        Cluster labels in the order of samples in ``X``.
    centers_ : array-like, shape (n_clusters, n_features)
        Average scale values corresponding to each cluster.
    labels_centers_ : array-like, shape (n_clusters)
        Cluster labels for each cluster center.
    medoids_ : array-like, shape (n_clusters, n_features)
        Representative samples, one for each cluster.
    labels_medoids_ :  array-like, shape (n_clusters)
        Cluster labels for each medoid.
    is_medoid_ : array-like, shape (n_samples)
        Array indicating samples being medoids (1) or not (0). Same order as ``labels_``.
    medoid_names_ : list
        Names of the medoids. Set if ``names`` is provided to ``.fit``.
    """
    def __init__(self,
                 model_class: Type[ClusterMixin] = KMeans,
                 model_kwargs: Optional[Dict] = None,
                 verbose: bool = True,
                 random_state: Optional[str] = None,
                 ):
        """
        Parameters
        ----------
        model_class : Type[ClusterMixin], default=KMeans
            A clustering model class with ``n_clusters`` parameter.
        model_kwargs : dict, optional
            Keyword arguments to pass to the selected clustering model.
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        random_state : int, optional
            The seed used by the random number generator. If a positive integer, results of stochastic processes are
            consistent, enabling reproducibility. If ``None``, stochastic processes will be truly random.

        Notes
        -----
        * All attributes are set during fitting via the :meth:`AAclust.fit` method  and can be directly accessed.
        * AAclust is designed primarily for amino acid scales but can be used for any set of numerical indices.

        See Also
        --------
        * :class:`AAclustPlot`: the respective plotting class.
        * Scikit-learn `clustering model classes <https://scikit-learn.org/stable/modules/clustering.html>`_.

        Examples
        --------
        .. include:: examples/aaclust.rst
        """
        # Global parameters
        verbose = ut.check_verbose(verbose)
        random_state = ut.check_random_state(random_state=random_state)
        # Model parameters
        ut.check_mode_class(model_class=model_class)
        model_kwargs = ut.check_model_kwargs(model_class=model_class,
                                             model_kwargs=model_kwargs,
                                             param_to_check="n_clusters",
                                             random_state=random_state)
        # Internal attributes
        self._verbose = verbose
        self._random_state = random_state
        self._model_class = model_class
        self._model_kwargs = model_kwargs
        # Output parameters (set during model fitting)
        self.model : Optional[ClusterMixin] = None
        self.n_clusters: Optional[int] = None
        self.labels_: Optional[ut.ArrayLike1D] = None
        self.centers_: Optional[ut.ArrayLike1D] = None
        self.labels_centers_: Optional[ut.ArrayLike1D] = None
        self.medoids_: Optional[ut.ArrayLike1D] = None
        self.labels_medoids_: Optional[ut.ArrayLike1D] = None
        self.is_medoid_: Optional[ut.ArrayLike1D] = None
        self.medoid_names_: Optional[List[str]] = None

    def fit(self,
            X: ut.ArrayLike2D,
            n_clusters: Optional[int] = None,
            on_center: bool = True,
            min_th: float = 0.3,
            merge: bool = True,
            metric: str = "euclidean",
            names: Optional[List[str]] = None
            ) -> "AAclust":
        """
        Applies AAclust algorithm to feature matrix (``X``).

        Introduced in [Breimann24a]_, AAclust determines the optimal number of clusters, k, without pre-specification.
        It partitions data (``X``) into clusters by maximizing the within-cluster Pearson correlation beyond the
        ``min_th`` threshold. The quality of clustering is either based on the minimum Pearson correlation of all
        members (``on_center=False``) or between the cluster center and its members (``on_center=True``),
        using either the ``min_cor_all`` or ``min_cor_center`` correlation measures, respectively.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        n_clusters : int, optional
            Pre-defined number of clusters. If provided, k is not optimized. Must be 0 > n_clusters > n_samples.
        min_th : float, default=0.3
            Pearson correlation threshold for clustering optimization (between 0 and 1).
        on_center : bool, default=True
            If ``True``, ``min_th`` is applied to the cluster center. Otherwise, to all cluster members.
        merge : bool, default=True
            If ``True``, the optional merging step is performed.
        metric : {'correlation', 'euclidean', 'manhattan', 'cosine'}, default='euclidean'
            Similarity measure used for optional cluster merging and obtaining medoids:

             - ``correlation``: Pearson correlation (maximum)
             - ``euclidean``: Euclidean distance (minimum)
             - ``manhattan``: Manhattan distance (minimum)
             - ``cosine``: Cosine distance (minimum)

        names : list of str, optional
            List of sample names. If provided, sets :attr:`AAclust.medoid_names_` attribute.

        Returns
        -------
        AAclust
            The fitted instance of the AAclust class, allowing direct attribute access.

        Notes
        -----
        * The ``AAclust`` algorithm consists of three main steps:

            1. Estimate the lower bound of k.
            2. Refine k (recursively) using the chosen quality measure.
            3. Optionally, merge smaller clusters as directed by the merge ``metric``.

        * ``AAclust`` provides two correlation-based quality measure to optimize ``n_clusters``:

            - ``min_cor_center``: Minimum Pearson correlation between the cluster center and all cluster members.
            - ``min_cor_all``: Minium pairwise Pearson correlation among all cluster members.

        * A representative scale (medoid) closest to each cluster center is selected for redundancy reduction.

        See Also
        --------
        * :func:`sklearn.metrics.pairwise_distances` were used as distances for merging.

        Warnings
        --------
        * All RuntimeWarnings during the AAclust algorithm are caught and bundled into one RuntimeWarning.

        Examples
        --------
        .. include:: examples/aaclust_fit.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        names = ut.check_list_like(name="names", val=names, accept_none=True)
        ut.check_number_range(name="min_th", val=min_th, min_val=0, max_val=1, just_int=False)
        ut.check_number_range(name="n_clusters", val=n_clusters, min_val=1, just_int=True, accept_none=True)
        check_metric(metric=metric)
        ut.check_bool(name="on_center", val=on_center)
        check_match_X_n_clusters(X=X, n_clusters=n_clusters, accept_none=True)
        check_match_X_names(X=X, names=names, accept_none=True)
        args = dict(model=self._model_class, model_kwargs=self._model_kwargs, min_th=min_th, on_center=on_center)
        # Clustering using given clustering models
        if n_clusters is not None:
            self.model = self._model_class(n_clusters=n_clusters, **self._model_kwargs)
            labels = self.model.fit(X).labels_.tolist()

        # Clustering using AAclust algorithm
        else:
            # 1. Step: Estimation of lower bound of k (number of clusters)
            if self._verbose:
                ut.print_out("1. Estimation of lower bound of k (number of clusters)", end="")
            n_clusters_lb = estimate_lower_bound_n_clusters(X, **args)
            # 2. Step: Optimization of k by recursive clustering
            if self._verbose:
                objective_fct = "min_cor_center" if on_center else "min_cor_all"
                ut.print_out(f"2. Optimization of k by recursive clustering ({objective_fct}, min_th={min_th}, k={n_clusters_lb})", end="")
            n_clusters = optimize_n_clusters(X, n_clusters=n_clusters_lb, **args)
            self.model = self._model_class(n_clusters=n_clusters, **self._model_kwargs)
            labels = self.model.fit(X).labels_.tolist()
            # 3. Step: Cluster merging (optional)
            if metric is not None:
                if self._verbose:
                    ut.print_out(f"3. Cluster merging (k={len(labels)})", end="")
                labels = merge_clusters(X, labels=labels, min_th=min_th, on_center=on_center, metric=metric)
                n_clusters = len(set(labels))

        # Obtain cluster centers and medoids
        medoids, medoid_labels, medoid_ind = compute_medoids(X, labels=labels)
        centers, center_labels = compute_centers(X, labels=labels)

        # Save results in output parameters
        post_check_n_clusters(n_clusters_actual=len(set(labels)), n_clusters=n_clusters)
        self.n_clusters = len(set(labels))
        self.labels_ = np.array(labels)
        self.centers_ = centers
        self.labels_centers_ = center_labels
        self.medoids_ = medoids     # Representative scales
        self.labels_medoids_ = medoid_labels
        self.is_medoid_ = np.array([i in medoid_ind for i in range(0, len(labels))])
        if names is not None:
            self.medoid_names_ = [names[i] for i in medoid_ind]
        return self

    def eval(self,
             X: ut.ArrayLike2D,
             list_labels: ut.ArrayLike2D = None,
             names_datasets: Optional[List[str]] = None,
             ) -> pd.DataFrame:
        """
        Evaluates the quality of different clustering results.

        The following established clustering measures are used:

        - ``BIC`` (Bayesian Information Criterion): Reflects the goodness of fit for the clustering while accounting for
          the number of clusters and parameters. The BIC value can range from negative infinity to positive infinity.
          A higher BIC indicates superior clustering quality.
        - ``CH`` (Calinski-Harabasz Index): Represents the ratio of between-cluster dispersion mean to the within-cluster dispersion.
          The CH value ranges from 0 to positive infinity. A higher CH score suggests better-defined clustering.
        - ``SC`` (Silhouette Coefficient): Evaluates the proximity of each data point in one cluster to the points in the neighboring clusters.
          The SC score lies between -1 and 1. A value closer to 1 implies better clustering.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        list_labels : array-like, shape (n_datasets, n_samples)
            List of arrays with cluster labels for samples in ``X`` obtained by the :meth:`AAclust.fit` method.
            Unique label values indicate clusters.
        names_datasets : list, optional
            List of dataset names corresponding to ``list_labels``.

        Returns
        -------
        df_eval : pd.DataFrame
            Evaluation results for each set of clustering labels from ``list_labels``.

        Notes
        -----
        ``df_eval`` includes the following columns:

            - 'names': Names (string) of evaluated datasets.
            - 'n_clusters': Number (integer) of clusters, equal to number of medoids.
            - 'BIC': BIC value (float) for clustering (-inf to inf).
            - 'CH': CH value (float) for clustering (0 to inf).
            - 'SC': SC value (float) for clustering (-1 to 1).

        BIC was adapted form this `StackExchange discussion <https://stats.stackexchange.com/questions/90769/using-bic-to-estimate-the-number-of-k-in-kmeans>`_
        and modified to align with the SC and CH score so that higher values signify better clustering,
        contrary to conventional BIC implementation favoring lower values. See [Breimann24a]_.

        See Also
        --------
        * :meth:`AAclustPlot.eval`: the respective plotting method.
        * :func:`sklearn.metrics.silhouette_score`: a commonly used clustering quality measures.
        * :func:`sklearn.metrics.calinski_harabasz_score`: a commonly used clustering quality measures.

        Examples
        --------
        .. include:: examples/aaclust_eval.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        list_labels = ut.check_array_like(name="list_labels", val=list_labels, ensure_2d=True, convert_2d=True)
        ut.check_match_X_list_labels(X=X, list_labels=list_labels)
        names_datasets = ut.check_list_like(name="names_datasets", val=names_datasets, accept_none=True, accept_str=True,
                                            check_all_str_or_convertible=True)
        ut.check_match_list_labels_names_datasets(list_labels=list_labels, names_datasets=names_datasets)
        # Get number of clusters (number of medoids) and evaluation measures
        df_eval, warn_ch, warn_sc = evaluate_clustering(X, list_labels=list_labels, names_datasets=names_datasets)
        if warn_ch and self._verbose:
            warnings.warn("CH was set to 0 because sklearn.metric.calinski_harabasz_score returned NaN.", RuntimeWarning)
        if warn_sc and self._verbose:
            warnings.warn("SC was set to -1 because sklearn.metric.silhouette_score returned NaN.", RuntimeWarning)
        return df_eval

    @staticmethod
    def name_clusters(X: ut.ArrayLike2D,
                      labels: ut.ArrayLike1D = None,
                      names: List[str] = None,
                      shorten_names : bool = True,
                      ) -> List[str]:
        """
        Assigns names to clusters based on the frequency of names.

        Names with higher frequency are prioritized. If a name is already assigned to a cluster,
        or the cluster contains one sample, its name is set to 'Unclassified'.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        labels : array-like, shape (n_samples,)
            Cluster labels for each sample in ``X``.
        names : list of str
            List of sample names corresponding to ``X``.
        shorten_names : bool, default=True
            If ``True``, shorten version of the names will be used.

        Returns
        -------
        cluster_names : list of str
            A list of renamed clusters based on names.

        Examples
        --------
        .. include:: examples/aaclust_name_clusters.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        labels = ut.check_labels(labels=labels)
        names = ut.check_list_like(name="names", val=names, accept_none=False)
        ut.check_bool(name="shorten_names", val=shorten_names)
        ut.check_match_X_labels(X=X, labels=labels)
        check_match_X_names(X=X, names=names, accept_none=False)
        # Get cluster names
        cluster_names = name_clusters(X, labels=labels, names=names, shorten_names=shorten_names)
        return cluster_names

    @staticmethod
    def comp_centers(X: ut.ArrayLike2D,
                     labels: ut.ArrayLike1D = None
                     ) -> Tuple[ut.ArrayLike1D, ut.ArrayLike1D]:
        """
        Computes the center of each cluster based on the given labels.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        labels : array-like, shape (n_samples,)
            Cluster labels for each sample in ``X``.

        Returns
        -------
        centers : array-like, shape (n_clusters,)
            The computed center for each cluster.
        labels_centers : array-like, shape (n_clusters,)
            The labels associated with each computed center.

        Examples
        --------
        .. include:: examples/aaclust_comp_centers.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        labels = ut.check_labels(labels=labels)
        ut.check_match_X_labels(X=X, labels=labels)
        # Get cluster centers
        centers, labels_centers = compute_centers(X, labels=labels)
        return centers, labels_centers

    @staticmethod
    def comp_medoids(X: ut.ArrayLike2D,
                     labels: ut.ArrayLike1D = None,
                     metric: str = "correlation"
                     ) -> Tuple[ut.ArrayLike1D, ut.ArrayLike1D]:
        """
        Computes the medoid of each cluster based on the given labels.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        labels : array-like, shape (n_samples,)
            Cluster labels for each sample in ``X``.
        metric : {'correlation', 'euclidean', 'manhattan', 'cosine'}, default='correlation'
            Similarity measure used to obtain medoids:

             - ``correlation``: Pearson correlation (maximum)
             - ``euclidean``: Euclidean distance (minimum)
             - ``manhattan``: Manhattan distance (minimum)
             - ``cosine``: Cosine distance (minimum)

        Returns
        -------
        medoids : array-like, shape (n_clusters,)
            The medoid for each cluster.
        labels_medoids : array-like, shape (n_clusters,)
            The labels corresponding to each medoid.

        Examples
        --------
        .. include:: examples/aaclust_comp_medoids.rst
        """
        # Check input
        X = ut.check_X(X=X)
        ut.check_X_unique_samples(X=X)
        labels = ut.check_labels(labels=labels)
        ut.check_match_X_labels(X=X, labels=labels)
        check_metric(metric=metric)
        # Get cluster medoids
        medoids, labels_medoids, _ = compute_medoids(X, labels=labels, metric=metric)
        return medoids, labels_medoids

    @staticmethod
    def comp_correlation(X: ut.ArrayLike2D,
                         labels: ut.ArrayLike1D = None,
                         X_ref: Optional[ut.ArrayLike2D] = None,
                         labels_ref: Optional[ut.ArrayLike1D] = None,
                         names : Optional[List[str]] = None,
                         names_ref : Optional[List[str]] = None
                         ) -> Tuple[pd.DataFrame, ut.ArrayLike1D]:
        """
        Computes the Pearson correlation of given data with reference data.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Feature matrix. `Rows` typically correspond to scales and `columns` to amino acids.
        labels : array-like, shape (n_samples,)
            Cluster labels for each sample in ``X``.
        X_ref : array-like, shape (n_samples, n_features)
            Feature matrix of reference data. If given, samples of ``X`` are compared with samples of ``X_ref``.
        labels_ref  : array-like, shape (n_samples_ref,)
            Cluster labels for each sample in ``X_ref``.
        names : list of str, optional
            List of sample names corresponding to ``X``.
        names_ref : list of str, optional
            List of sample names corresponding to ``X_ref``.

        Returns
        -------
        df_corr : pd.DataFrame, shape (n_samples, n_samples or n_samples_ref)
            DataFrame with correlation either for each pair in ``X`` of shape (n_samples, n_samples) or
            for each pair between ``X`` and ``X_ref`` of shape (n_samples, n_samples_ref).
        labels_sorted: array-like, shape (n_samples_ref,)
            Cluster labels for each sample and sorted as in `df_corr`.

        Notes
        -----
        * Rows will be sorted in ascending order of ``labels``.
        * Columns will be sorted in ascending order of ``labels`` or ``labels_ref`` if given.
        * Labels are replaced by respective names if given.

        See Also
        --------
        * :meth:`pandas.DataFrame.corr` used to compute the correlation.

        Examples
        --------
        .. include:: examples/aaclust_comp_correlation.rst
        """
        # Check input
        X = ut.check_X(X=X, min_n_samples=2)
        ut.check_X_unique_samples(X=X, min_n_unique_samples=2)
        labels = check_labels_cor(labels=labels, labels_name="labels")
        ut.check_match_X_labels(X=X, labels=labels)
        check_match_X_names(X=X, names=names, accept_none=True)
        if X_ref is not None:
            X_ref = ut.check_X(X=X_ref, min_n_samples=1)
            labels_ref = check_labels_cor(labels=labels_ref, labels_name="labels_ref")
            ut.check_match_X_labels(X=X_ref, labels=labels_ref)
            check_match_X_names(X=X_ref, names=names_ref, accept_none=True)
            check_X_X_ref(X=X, X_ref=X_ref)
        # Get correlations
        df_corr, labels_sorted = compute_correlation(X, X_ref=X_ref,
                                                     labels=labels, labels_ref=labels_ref,
                                                     names=names, names_ref=names_ref)
        return df_corr, labels_sorted

    @staticmethod
    def comp_coverage(names : List[str] = None,
                      names_ref : List[str] = None
                      ) -> float :
        """
        Computes the percentage of unique names from ``names`` that are present in ``names_ref``.

        This method helps in understanding the coverage of a particular set of names (subset)
        within a reference set of names (universal set). Each name from both ``names`` and ``names_ref``
        are considered only once, regardless of repetition.

        Parameters
        ----------
        names, list of str
            List of sample names. Should be subset of ``names_ref``.
        names_ref, list of str
            List of reference sample names. Should be superset of ``names``.

        Returns
        -------
        coverage : float
            Percentage of unique names from ``names`` that are found in ``names_ref``.

        Examples
        --------
        .. include:: examples/aaclust_comp_coverage.rst
        """
        names = ut.check_list_like(name="names", val=names, accept_none=False)
        names_ref = ut.check_list_like(name="names_ref", val=names_ref, accept_none=False)
        ut.check_superset_subset(subset=names, name_subset="names",
                                 superset=names_ref, name_superset="names_ref")
        # Compute coverage
        n_unique_names = len(set(names))
        n_unique_ref = len(set(names_ref))
        coverage = round(n_unique_names/n_unique_ref*100, 2)
        return coverage



"""
This is a script for the frontend of the SeqMutPlot class
"""
import aaanalysis.utils as ut


# I Helper Functions
# Check functions


# II Main Functions
class SeqMutPlot:
    """
    UNDER CONSTRUCTION - Plotting class for ``SeqMut`` (Sequence Mutator).
    """
    def __init__(self, verbose=False, df_scales=None):
        self._verbose = ut.check_verbose(verbose)
        self.df_scales = df_scales

    # Main method
    def mutation_landscape(self):
        """"""


"""
This is a script for the backend of the CPPPlot.heatmap() method.
"""
import matplotlib.pyplot as plt

import aaanalysis.utils as ut
from ._utils_cpp_plot_elements import PlotElements
from ._utils_cpp_plot_map import plot_heatmap_


# II Main Functions
def plot_heatmap(df_feat=None, df_cat=None,
                 shap_plot=False,
                 col_cat="subcategory", col_val="mean_dif",
                 name_test="TEST", name_ref="REF",
                 figsize=(8, 8),
                 start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10,
                 tmd_seq=None, jmd_n_seq=None, jmd_c_seq=None,
                 tmd_color="mediumspringgreen", jmd_color="blue",
                 tmd_seq_color="black", jmd_seq_color="white",
                 seq_size=None, fontsize_tmd_jmd=None, weight_tmd_jmd="normal",
                 fontsize_labels=None,
                 add_xticks_pos=False,
                 grid_linewidth=0.01, grid_linecolor=None,
                 border_linewidth=2,
                 facecolor_dark=None, vmin=None, vmax=None,
                 cmap=None, cmap_n_colors=101,
                 cbar_pct=True, cbar_kws=None, cbar_xywh=(0.7, None, 0.2, None),
                 dict_color=None, legend_kws=None, legend_xy=(-0.1, -0.01),
                 xtick_size=11.0, xtick_width=2.0, xtick_length=5.0):
    """Plot heatmap of feature values"""
    # Set fontsize
    pe = PlotElements()
    fs = ut.plot_gco()
    fs_labels = fs if fontsize_labels is None else fontsize_labels

    # Group arguments
    args_len = dict(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
    args_seq = dict(jmd_n_seq=jmd_n_seq, tmd_seq=tmd_seq, jmd_c_seq=jmd_c_seq)
    args_part_color = dict(tmd_color=tmd_color, jmd_color=jmd_color)
    args_seq_color = dict(tmd_seq_color=tmd_seq_color, jmd_seq_color=jmd_seq_color)
    args_fs = dict(seq_size=seq_size, fontsize_tmd_jmd=fontsize_tmd_jmd, fontsize_labels=fs_labels)
    args_xtick = dict(xtick_size=xtick_size, xtick_width=xtick_width, xtick_length=xtick_length)

    # Set SHAP arguments
    if facecolor_dark is None:
        facecolor_dark = shap_plot and ut.COL_FEAT_IMPACT in col_val

    if shap_plot and ut.COL_FEAT_IMPACT in col_val:
        cmap = ut.STR_CMAP_SHAP if cmap is None else cmap
        label_cbar = ut.LABEL_CBAR_FEAT_IMPACT_CUM
    else:
        cmap = ut.STR_CMAP_CPP if cmap is None else cmap
        label_cbar = f"Feature value\n{name_test} - {name_ref}"

    # Plot
    fig, ax = plt.subplots(figsize=figsize)

    # Set colorbar and legend arguments
    _cbar_kws, cbar_ax = pe.adjust_cbar_kws(fig=fig,
                                            cbar_kws=cbar_kws,
                                            cbar_xywh=cbar_xywh,
                                            label=label_cbar,
                                            fontsize_labels=fs_labels)

    # Set cat legend arguments
    n_cat = len(set(df_feat[ut.COL_CAT]))
    _legend_kws = pe.adjust_cat_legend_kws(legend_kws=legend_kws,
                                           n_cat=n_cat,
                                           legend_xy=legend_xy,
                                           fontsize_labels=fs_labels)

    # Plot heatmap
    ax = plot_heatmap_(df_feat=df_feat, df_cat=df_cat,
                       col_cat=col_cat, col_val=col_val,
                       ax=ax, figsize=figsize,
                       start=start, **args_len, **args_seq,
                       **args_part_color, **args_seq_color,
                       **args_fs, weight_tmd_jmd=weight_tmd_jmd,
                       add_xticks_pos=add_xticks_pos,
                       grid_linecolor=grid_linecolor, grid_linewidth=grid_linewidth,
                       border_linewidth=border_linewidth,
                       facecolor_dark=facecolor_dark, vmin=vmin, vmax=vmax,
                       cmap=cmap, cmap_n_colors=cmap_n_colors,
                       cbar_ax=cbar_ax, cbar_pct=cbar_pct, cbar_kws=_cbar_kws,
                       dict_color=dict_color, legend_kws=_legend_kws,
                       **args_xtick)
    return fig, ax



"""
This is a script for the backend of the ShapExplainer.add_feat_impact() and ShapExplainer.add_feat_importance() methods.
"""
import numpy as np
import pandas as pd
import warnings

import aaanalysis.utils as ut


# I Helper Functions
def _abs_normalize(values=None):
    total_impact = sum(np.abs(values))
    values = [round(x / total_impact * 100, 2) for x in values]
    return values


# II Main Functions
def _drop_feat_columns(df_feat=None):
    """Drop all feature impact and feature importance columns"""
    f = lambda x: x not in [ut.COL_FEAT_IMPORT, ut.COL_FEAT_IMPORT_STD] and ut.COL_FEAT_IMPACT not in x
    columns = [x for x in list(df_feat) if f(x)]
    df_feat = df_feat[columns]
    return df_feat

def _comp_sample_shap_feat_impact(shap_values=None, i=None, normalize=True):
    """Compute the shap feature impact for the i-th sample"""
    shap_value_sample = shap_values[i]
    feat_impact = shap_value_sample / np.nansum(np.abs(shap_value_sample), axis=0) * 100
    if normalize:
        feat_impact = _abs_normalize(values=feat_impact)
    return feat_impact


def _comp_group_shap_feat_impact(shap_values=None, list_i=None, normalize=True, verbose=True):
    """Compute the shap feature impact for a group of samples"""
    shap_value_samples = shap_values[list_i]
    mean_shap_value_samples = shap_value_samples.mean(axis=0)
    std_shap_values_samples = shap_value_samples.std(axis=0)
    feat_impact = mean_shap_value_samples / np.nansum(np.abs(mean_shap_value_samples)) * 100
    feat_impact_std = std_shap_values_samples / np.nansum(np.abs(mean_shap_value_samples)) * 100
    if normalize:
        abs_sum_before_norm = sum(np.abs(feat_impact))
        feat_impact = _abs_normalize(values=feat_impact)
        abs_sum_after_norm = sum(np.abs(feat_impact))
        # Scale std (assumes linear scaling factor)
        feat_impact_std *= abs_sum_after_norm/abs_sum_before_norm
    if verbose:
        max_std, max_impact = round(max(np.abs(feat_impact_std)), 2), round(max(np.abs(feat_impact)), 2)
        if max_std > max_impact * 5:
            warnings.warn(f"Absolute maximum of 'feat_impact_std' ({max_std}) >> 'feat_impact' ({max_impact}). Grouping might be invalid.")
    return feat_impact, feat_impact_std


# II Main functions
# Feature importance
def comp_shap_feature_importance(shap_values=None, normalize=True):
    """
    Compute shap feature importance (absolute mean for all samples)
    See: https://github.com/slundberg/shap/issues/538
    """
    feat_importance = np.abs(shap_values).mean(axis=0) * 100
    if normalize:
        feat_importance = _abs_normalize(values=feat_importance)
    return feat_importance


def insert_shap_feature_importance(df_feat=None, feat_importance=None, drop=False):
    """Insert shap explainer-based feature importance"""
    df_feat = df_feat.copy()
    if drop:
        df_feat = _drop_feat_columns(df_feat=df_feat)
    args = dict(allow_duplicates=False)
    df_feat.insert(loc=len(df_feat.columns), column=ut.COL_FEAT_IMPORT, value=feat_importance, **args)
    return df_feat


# Feature impact
def comp_shap_feature_impact(shap_values, sample_positions=None, normalize=True, group_average=False, verbose=True):
    """
    Compute SHAP feature impact for different scenarios:
        a) For a single sample, returning its feature impact.
        b) For multiple samples, returning each sample's feature impact.
        c) For a group of samples, returning the group average feature impact.
    """
    # Single sample
    if isinstance(sample_positions, int):
        return _comp_sample_shap_feat_impact(shap_values, i=sample_positions, normalize=normalize)
    # Multiple samples
    elif isinstance(sample_positions, list) and not group_average:
        impacts = [_comp_sample_shap_feat_impact(shap_values, i, normalize) for i in sample_positions]
        return np.array(impacts)
    # Group average
    elif isinstance(sample_positions, list) and group_average:
        feat_impact, feat_impact_std = _comp_group_shap_feat_impact(shap_values, list_i=sample_positions, normalize=normalize,
                                                                    verbose=verbose)
        return np.array([feat_impact, feat_impact_std])


def insert_shap_feature_impact(df_feat=None, feat_impact=None, names=None, group_average=False, drop=False):
    """Insert shap explainer-based feature importance"""
    df_feat = df_feat.copy()
    if drop:
        df_feat = _drop_feat_columns(df_feat=df_feat)
    # Single sample or multiple samples
    if not group_average:
        col_names = [f'{ut.COL_FEAT_IMPACT}_{col_name}' for col_name in names]
    # Group average
    else:
        col_names = [f'{ut.COL_FEAT_IMPACT}_{names}', f'{ut.COL_FEAT_IMPACT_STD}_{names}']
    df_feat_impact = pd.DataFrame(data=feat_impact.T, columns=col_names)
    df_feat = pd.concat([df_feat, df_feat_impact], axis=1)
    return df_feat



"""
This is a script for utility feature functions for CPP and SequenceFeature objects and backend.
"""
import os
import numpy as np
from itertools import repeat
import multiprocessing as mp
import pandas as pd

from ._part import create_parts
from ._split import Split
import aaanalysis.utils as ut


# I Helper Functions
# DEV: Run exceptionally in backend because error might only be caused internally
def pre_check_vf_scale(part_split=None):
    """Check for the presence of a specific gap string in the sequence."""
    # Join the elements of part_split into a single string
    combined_string = "".join(part_split)
    # Check if ut.STR_AA_GAP is in the combined string
    if ut.STR_AA_GAP in combined_string:
        raise ValueError("Some input sequences contain gaps ('-').")

def post_check_vf_scale(feature_values=None):
    """Check if feature_values/X does contain nans due to gaps"""
    if np.isnan(feature_values).any():
        raise ValueError("Some input sequences result in NaN feature values most likely due to gaps ('-').")


def get_vf_scale(dict_scale=None, accept_gaps=False):
    """Vectorized function to calculate the mean for a feature"""
    if not accept_gaps:
        # Vectorized scale function
        vf_scale = np.vectorize(lambda x: np.mean([dict_scale[a] for a in x]))
    else:
        # Except NaN derived from 'X' in sequence if not just 'X' in sequence (3x slower)
        def get_mean_excepting_nan(x):
            vals = np.array([dict_scale.get(a, np.NaN) for a in x])
            return vals
        vf_scale = np.vectorize(lambda x: np.nanmean(get_mean_excepting_nan(x)))
    return vf_scale


def _get_split_info(split=None):
    """Parse split information"""
    split = split.replace(" ", "")  # remove whitespace
    # Check Segment
    if ut.STR_SEGMENT in split:
        split_type = ut.STR_SEGMENT
        i_th, n_split = [int(x) for x in split.split("(")[1].replace(")", "").split(",")]
        split_kwargs = dict(i_th=i_th, n_split=n_split)
    # Check PeriodicPattern
    elif ut.STR_PERIODIC_PATTERN in split:
        split_type = ut.STR_PERIODIC_PATTERN
        start = split.split("i+")[1].replace(")", "").split(",")
        step1, step2 = [int(x) for x in start.pop(0).split("/")]
        start = int(start[0])
        terminus = split.split("i+")[0].split("(")[1].replace(",", "")
        split_kwargs = dict(terminus=terminus, step1=step1, step2=step2, start=start)
    # Check pattern
    else:
        split_type = ut.STR_PATTERN
        list_pos = split.split("(")[1].replace(")", "").split(",")
        terminus = list_pos.pop(0)
        list_pos = [int(x) for x in list_pos]
        split_kwargs = dict(terminus=terminus, list_pos=list_pos)
    return split_type, split_kwargs


# Get positions
def _get_dict_part_pos(tmd_len=20, jmd_n_len=10, jmd_c_len=10, start=1):
    """Get dictionary for part to positions."""
    jmd_n = list(range(0, jmd_n_len))
    tmd = list(range(jmd_n_len, tmd_len+jmd_n_len))
    jmd_c = list(range(jmd_n_len + tmd_len, jmd_n_len + tmd_len + jmd_c_len))
    # Change int to string and adjust length
    jmd_n = [i + start for i in jmd_n]
    tmd = [i + start for i in tmd]
    jmd_c = [i + start for i in jmd_c]
    dict_part_pos = ut.get_dict_part_seq(tmd=tmd, jmd_n=jmd_n, jmd_c=jmd_c)
    return dict_part_pos


def _get_positions(dict_part_pos=None, features=None, as_str=True):
    """Get list of positions for given feature names."""
    sp = Split(type_str=False)
    list_pos = []
    for feat_id in features:
        part, split, scale = feat_id.split("-")
        split_type, split_kwargs = _get_split_info(split=split)
        f_split = getattr(sp, split_type.lower())
        pos = sorted(f_split(seq=dict_part_pos[part.lower()], **split_kwargs))
        if as_str:
            pos = str(pos).replace("[", "").replace("]", "").replace(" ", "")
        list_pos.append(pos)
    return list_pos


# Get df positions
def _get_df_pos_long(df=None, col_cat="category", col_val=None):
    """Get """
    if col_val is None:
        df_feat = df[[ut.COL_FEATURE, col_cat]].set_index(ut.COL_FEATURE)
    else:
        df_feat = df[[ut.COL_FEATURE, col_cat, col_val]].set_index(ut.COL_FEATURE)
    # Columns = scale categories, rows = features
    df_pos_long = pd.DataFrame(df[ut.COL_POSITION].str.split(",").tolist())
    df_pos_long.index = df[ut.COL_FEATURE]
    df_pos_long = df_pos_long.stack().reset_index(level=1).drop("level_1", axis=1).rename({0: ut.COL_POSITION}, axis=1)
    df_pos_long = df_pos_long.join(df_feat)
    df_pos_long[ut.COL_POSITION] = df_pos_long[ut.COL_POSITION].astype(int)
    return df_pos_long


# Get feature matrix
def _get_dict_all_scales(df_scales=None):
    """Get nested dictionary where each scale is a key for an amino acid scale value dictionary"""
    dict_all_scales = {col: dict(zip(df_scales.index.to_list(), df_scales[col])) for col in list(df_scales)}
    return dict_all_scales


def _feature_value(df_parts=None, split=None, dict_scale=None, accept_gaps=False):
    """Helper function to create feature values for feature matrix"""
    sp = Split()
    # Get vectorized split function
    split_type, split_kwargs = _get_split_info(split=split)
    f_split = getattr(sp, split_type.lower())
    # Vectorize split function using anonymous function
    vf_split = np.vectorize(lambda x: f_split(seq=x, **split_kwargs))
    # Get vectorized scale function
    vf_scale = get_vf_scale(dict_scale=dict_scale, accept_gaps=accept_gaps)
    # Combine part split and scale to get feature values
    part_split = vf_split(df_parts)
    if not accept_gaps:
        pre_check_vf_scale(part_split=part_split)
    feature_value = np.round(vf_scale(part_split), 5)  # feature values
    if accept_gaps:
        post_check_vf_scale(feature_values=feature_value)
    return feature_value


def _feature_matrix(features, dict_all_scales, df_parts, accept_gaps):
    """Helper function to create feature matrix via multiple processing"""
    X = np.empty([len(df_parts), len(features)])
    for i, feat_name in enumerate(features):
        part, split, scale = feat_name.split("-")
        dict_scale = dict_all_scales[scale]
        X[:, i] = _feature_value(split=split,
                                 dict_scale=dict_scale,
                                 df_parts=df_parts[part.lower()],
                                 accept_gaps=accept_gaps)
    return X


# II Main Functions
def get_part_positions(start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10):
    """Get part positions"""
    jmd_n = list(range(start, jmd_n_len + start))
    tmd = list(range(jmd_n_len + start, jmd_n_len + tmd_len + start))
    jmd_c = list(range(jmd_n_len + tmd_len + start, jmd_n_len + tmd_len + jmd_c_len + start))
    return jmd_n, tmd, jmd_c


def get_list_parts(features=None):
    """Get list of parts to cover all features"""
    features = [features] if type(features) is str else features
    # Features are PART-SPLIT-SCALE combinations
    list_parts = list(set([x.split("-")[0].lower() for x in features]))
    return list_parts


def get_df_parts_(df_seq=None, list_parts=None, jmd_n_len=None, jmd_c_len=None):
    """Create DataFrame with sequence parts"""
    pos_based = set(ut.COLS_SEQ_POS).issubset(set(df_seq))
    dict_parts = {}
    for i, row in df_seq.iterrows():
        entry = row[ut.COL_ENTRY]
        if jmd_c_len is not None and jmd_n_len is not None and pos_based:
            seq, tmd_start, tmd_stop = row[ut.COLS_SEQ_POS].values
            jmd_n, tmd, jmd_c = create_parts(seq=seq, tmd_start=tmd_start, tmd_stop=tmd_stop,
                                             jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        else:
            jmd_n, tmd, jmd_c = row[ut.COLS_SEQ_PARTS].values
        dict_part_seq = ut.get_dict_part_seq(tmd=tmd, jmd_n=jmd_n, jmd_c=jmd_c)
        dict_part_seq = {part: dict_part_seq[part] for part in list_parts}
        dict_parts[entry] = dict_part_seq
    df_parts = pd.DataFrame.from_dict(dict_parts).T
    # DEV: the following line sorts index if list_parts contains just one element
    # df_parts = pd.DataFrame.from_dict(dict_parts, orient="index")
    return df_parts


def remove_entries_with_gaps_(df_parts=None):
    """Remove rows where any cell contains '-'"""
    df_parts = df_parts[~df_parts.map(lambda x: '-' in str(x)).any(axis=1)]
    return df_parts


def get_positions_(features=None, start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10):
    """Create list with positions for given feature names"""
    features = [features] if type(features) is str else features
    dict_part_pos = _get_dict_part_pos(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len, start=start)
    feat_positions = _get_positions(dict_part_pos=dict_part_pos, features=features)
    return feat_positions


def get_amino_acids_(features=None, tmd_seq="", jmd_n_seq="", jmd_c_seq=""):
    """Create amino acid segments/patterns for features"""
    features = [features] if type(features) is str else features
    pos = get_positions_(features=features, tmd_len=len(tmd_seq), jmd_n_len=len(jmd_n_seq),
                         jmd_c_len=len(jmd_c_seq), start=0)
    seq = jmd_n_seq + tmd_seq + jmd_c_seq
    f_seg = lambda x: "".join([seq[int(p)] for p in x.split(",")])
    f_pat = lambda x: "-".join([seq[int(p)] for p in x.split(",")])
    feat_aa = [f_seg(pos) if "Segment" in feat else f_pat(pos) for feat, pos in zip(features, pos)]
    return feat_aa


def get_feature_matrix_(features=None, df_parts=None, df_scales=None, accept_gaps=False, n_jobs=None):
    """Create feature matrix for given feature ids and sequence parts."""
    features = [features] if type(features) is str else features
    dict_all_scales = _get_dict_all_scales(df_scales=df_scales)
    # Convert features to list if needed
    features = features.to_list() if isinstance(features, pd.Series) else features
    # Feature creation
    if n_jobs == 1:
        # Run in a single process
        feat_matrix = _feature_matrix(features, dict_all_scales, df_parts, accept_gaps)
    else:
        # Run in multiple processes
        if n_jobs is None:
            # Optimize n_jobs that each core processes a minimum of 10 features
            n_jobs = min([os.cpu_count(), max([int(len(features)/10), 1])])
        feat_chunks = np.array_split(features, n_jobs)
        args = zip(feat_chunks, repeat(dict_all_scales), repeat(df_parts), repeat(accept_gaps))
        with mp.get_context("spawn").Pool(processes=n_jobs) as pool:
            result = pool.starmap(_feature_matrix, args)
        feat_matrix = np.concatenate(result, axis=1)
    return feat_matrix


def get_df_pos_(df_feat=None, col_cat="category", col_val=None, value_type="count", start=None, stop=None):
    """Get df with aggregated values for each combination of column values and positions"""
    df_feat = df_feat.copy()
    list_y_cat = sorted(set(df_feat[col_cat]))
    if value_type != "mean" and col_val is not None:
        df_feat[col_val] = df_feat[col_val] / [len(x.split(",")) for x in df_feat[ut.COL_POSITION]]
    # Get df with features for each position
    df_pos_long = _get_df_pos_long(df=df_feat, col_cat=col_cat, col_val=col_val)
    # Get dict with values of categories for each position
    dict_pos_val = {p: [] for p in range(start, stop+1)}
    dict_cat_val = {c: 0 for c in list_y_cat}
    for p in dict_pos_val:
        if value_type == "count":
            dict_val = dict(df_pos_long[df_pos_long[ut.COL_POSITION] == p][col_cat].value_counts())
        elif value_type == "mean":
            dict_val = dict(df_pos_long[df_pos_long[ut.COL_POSITION] == p].groupby(col_cat).mean()[col_val])
        elif value_type == "sum":
            dict_val = dict(df_pos_long[df_pos_long[ut.COL_POSITION] == p].groupby(col_cat).sum()[col_val])
        else:
            dict_val = dict(df_pos_long[df_pos_long[ut.COL_POSITION] == p].groupby(col_cat).std()[col_val])
        dict_pos_val[p] = {**dict_cat_val, **dict_val}
    # Get df with values (e.g., counts) of each category and each position
    df_pos = pd.DataFrame(dict_pos_val)
    df_pos = df_pos.T[list_y_cat].T     # Filter and order categories
    return df_pos


def get_df_pos_parts_(df_pos=None, value_type="sum", start=1, tmd_len=20, jmd_n_len=10, jmd_c_len=10, list_parts=None):
    """Get df with aggregated values for each combination of column values and sequence parts."""
    jmd_n, tmd, jmd_c = get_part_positions(start=start, jmd_n_len=jmd_n_len, tmd_len=tmd_len, jmd_c_len=jmd_c_len)
    dict_part_pos = ut.get_dict_part_seq(tmd=tmd, jmd_n=jmd_n, jmd_c=jmd_c)
    if value_type == "sum":
        list_df = [df_pos[dict_part_pos[part]].sum(axis=1) for part in list_parts]
    elif value_type == "mean":
        list_df = [df_pos[dict_part_pos[part]].mean(axis=1) for part in list_parts]
    elif value_type == "count":
        list_df = [df_pos[dict_part_pos[part]].value_counts() for part in list_parts]
    else:
        list_df = [df_pos[dict_part_pos[part]].std(axis=1) for part in list_parts]
    df_pos = pd.concat(list_df, axis=1)
    df_pos.columns = list_parts
    return df_pos


def add_scale_info_(df_feat=None, df_cat=None):
    """Add scale information to DataFrame (scale categories, sub categories, and scale names)."""
    # Add scale categories
    df_cat = df_cat.copy()
    i = df_feat.columns.get_loc(ut.COL_FEATURE)
    for col in [ut.COL_SCALE_DES, ut.COL_SCALE_NAME, ut.COL_SUBCAT, ut.COL_CAT]:
        if col in list(df_feat):
            df_feat.drop(col, inplace=True, axis=1)
        dict_cat = dict(zip(df_cat[ut.COL_SCALE_ID], df_cat[col]))
        vals = [dict_cat[s.split("-")[2]] for s in df_feat[ut.COL_FEATURE]]
        df_feat.insert(i + 1, col, vals)
    return df_feat


"""
This is a script for the frontend of the dPULearnPlot class for plotting results for the identification of
reliable negatives from unlabeled data using dPULearn.
"""
import time
import pandas as pd
import numpy as np
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from typing import Optional, Dict, Union, List, Tuple, Type


import aaanalysis.utils as ut
from ._backend.dpulearn.dpul_plot import (plot_eval, plot_pca)
# TODO visualize loadings as feature plot (long-term)


# I Helper Functions
def _adjust_plot_args(names=None, colors=None, args_scatter=None, default_names=None, default_colors=None):
    """Adjust names, colors, and args_scatter arguments"""
    # Set default names if not provided
    if names is None:
        names = default_names
    # Extract and adjust colors from args_scatter
    if args_scatter is not None and ("color" in args_scatter or "c" in args_scatter):
        if "color" in args_scatter:
            extracted_colors = args_scatter.pop('color', None)
        else:
            extracted_colors = args_scatter.pop("c", None)
        if isinstance(extracted_colors, str):
            colors = [extracted_colors] * len(names)
        elif extracted_colors is not None:
            colors = extracted_colors
    # Set default colors if not provided
    if colors is None:
        colors = default_colors
    # Validate the match between names and colors
    check_match_names_colors(names=names, colors=colors)
    # Adjust args_scatter with default and additional arguments
    _args_scatter = dict(linewidth=0.5, edgecolor="white")
    if args_scatter is not None:
        # Avoid aliases in args_scatter
        for arg in ["edgecolors", "linewidths"]:
            if arg in args_scatter:
                _args_scatter.pop(arg.replace("s", ""), None)
        _args_scatter.update(args_scatter)
    return names, colors, _args_scatter


# Check functions
def check_match_df_pu_labels(df_pu=None, labels=None):
    """Check length match df_pu and labels"""
    n_samples = len(df_pu)
    if n_samples != len(labels):
        raise ValueError(f"Number of samples from 'df_pu' (n={n_samples}) does not match with 'labels' (n={len(labels)})")


def check_match_names_colors(names=None, colors=None):
    """Check if length matches of names and colors"""
    if len(names) != len(colors):
        raise ValueError(f"Length of 'names' (n={len(names)}) and 'colors' (n={len(colors)}) does not match.")


# II Main Functions
class dPULearnPlot:
    """
    Plotting class for :class:`dPULearn` (deterministic Positive-Unlabeled Learning) results [Breimann24c]_.

    This plotting class visualizes the by ``dPULearn`` reliably identified negative samples in a compressed feature
    space.

    """
    def __int__(self):
        """
        See Also
        --------
        * :class:`dPULearn`: the respective plotting class.
        """

    @staticmethod
    def eval(df_eval: pd.DataFrame = None,
             figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
             dict_xlims: Optional[Union[None, dict]] = None,
             legend: bool = True,
             legend_y: float = -0.175,
             colors: Optional[List[str]] = None,
             ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot evaluation output of dPULearn comparing multiple sets of identified negatives.

        Evaluation measures can be grouped into 'Homogeneity' measures ('avg STD' and 'avg IQR') assessing
        the similarity within the sets of identified negatives, and 'Dissimilarity' measures ('avg AUC', 'avg KLD')
        assessing the dissimilarity between the identified negatives and the other reference groups including
        positive samples ('Pos'), unlabeled samples ('Unl'), and ground-truth negative samples ('Neg') if given.

        Parameters
        ----------
        df_eval : pd.DataFrame, shape (n_datasets, n_metrics)
            DataFrame with evaluation measures for sets of identified negatives. Each `row` corresponds to a specific
            dataset including identified negatives. Requiered 'columns' are:

            - 'name': Name of datasets containing identified negatives (typically named by identification approach).
            - 'avg_STD': Average standard deviation (STD) assessing homogeneity of identified negatives.
            - 'avg_IQR': Average interquartile range (IQR) assessing homogeneity of identified negatives.
            - 'avg_abs_AUC_DATASET': Average absolute area under the curve (AUC), which assesses the similarity between the
              set of identified negatives and other datasets. 'DATASET' must be 'pos' (positive samples) and 'unl'
              (unlabeled samples), as well as, optionally, 'neg' (ground-truth negative samples).

            Optional columns include:

            - 'avg_KLD_DATASET': The average Kullback-Leibler Divergence (KLD), which measures the distribution alignment
              between the set of identified negatives and other datasets ('pos', 'unl', or 'neg').

        figsize : tuple, default=(6, 4)
            Figure dimensions (width, height) in inches.
        dict_xlims : dict, optional
            A dictionary containing x-axis limits for subplots. Keys should be the subplot axis number ({0, 1, 2, 4})
            and values should be tuple specifying (``xmin``, ``xmax``). If ``None``, x-axis limits are auto-scaled.
        legend : bool, default=True
            If ``True``, legend is set under dissimilarity measures.
        legend_y : float, default=-0.175
            Legend position regarding the plot y-axis applied if ``legend=True``.
        colors : list of str, optional
            List of colors for identified negatives and the following reference datasets:
            positive samples ('Pos'), unlabeled samples ('Unl'), and ground-truth negative samples ('Neg').

        Returns
        -------
        fig : plt.Figure
            Figure object for evaluation plot
        axes : array of plt.Axes
            Array of Axes objects, each representing a subplot within the figure. .

        Notes
        -----
        * Ground-truth negatives are only shown if provided by ``df_eval``.

        See Also
        --------
        * :meth:`dPULearn.eval`: the respective computation method.
        * :func:`comp_auc_adjusted` and :func:`comp_kld`.

        Examples
        --------
        .. include:: examples/dpul_plot_eval.rst
        """
        # Check input
        cols_requiered = [ut.COL_NAME, ut.COL_AVG_STD, ut.COL_AVG_IQR, ut.COL_AVG_ABS_AUC_POS, ut.COL_AVG_ABS_AUC_UNL]
        ut.check_df(name="df_eval", df=df_eval, cols_requiered=cols_requiered, accept_none=False, accept_nan=False)
        ut.check_figsize(figsize=figsize, accept_none=True)
        ut.check_dict_xlims(dict_xlims=dict_xlims, n_ax=4)
        ut.check_bool(name="legend", val=legend)
        ut.check_number_val(name="legend_y", val=legend_y)
        ut.check_list_colors(name="colors", val=colors, accept_none=True, min_n=4)
        # Plotting
        if colors is None:
            colors = [ut.COLOR_REL_NEG, ut.COLOR_POS, ut.COLOR_UNL, ut.COLOR_NEG]
        fig, axes = plot_eval(df_eval=df_eval, colors=colors,
                              figsize=figsize, dict_xlims=dict_xlims,
                              legend=legend, legend_y=legend_y)
        return fig, axes

    @staticmethod
    def pca(df_pu: pd.DataFrame = None,
            labels=None,
            figsize: Tuple[Union[int, float], Union[int, float]] = (5, 5),
            pc_x: int = 1,
            pc_y: int = 2,
            show_pos_mean_x: bool = True,
            show_pos_mean_y: bool = True,
            colors: Optional[List[str]] = None,
            names: Optional[List[str]] = None,
            legend: bool = True,
            legend_y: float = -0.15,
            kwargs_scatterplot: Optional[dict] = None,
            ) -> plt.Axes:
        """
        Principal component analysis (PCA) plot for set of identified negatives.

        This method visualizes the differences between the set of identified negatives (labeled by 0) and the
        positive (1) and the unlabeled (2) sample groups. The selected principal components (PCs) represent
        a lower-dimensional feature space. Optionally, the average PC value for the positive samples can be shown,
        which was used for ``PCA-based identification`` of negatives.

        Parameters
        ----------
        df_pu : pd.DataFrame, shape (n_samples, pca_features)
            A DataFrame with the PCA-transformed features obtained from ``dPULearn.df_pu_``.
        figsize : tuple, default=(6, 6)
            Figure dimensions (width, height) in inches.
        labels : array-like, shape (n_samples,)
            Dataset labels of samples in ``df_pu``. Labels should contain 0 (identified negative) and 1 (positive).
            Unlabeled samples (2) can also be provided.
        pc_x : str, default='PC1'
            The name of the principal component (PC) to show at the x-axis.
        pc_y : str, default='PC2'
            The name of the principal component (PC) to show at the y-axis.
        show_pos_mean_x : bool, default False
            If ``True``, the mean of the x-axis PC values across the positive sample group is shown on the plot.
        show_pos_mean_y : bool, default False
            If ``True``, the mean of the y-axis PC values across the positive sample group is shown on the plot.
        colors : list of str, optional
            List of colors for identified negatives (0), positive samples (1), and unlabeled samples (2).
        names : list of str, optional
            List of dataset names for identified negatives, positive samples, and unlabeled samples.
        legend : bool, default=True
            If ``True``, legend is set under dissimilarity measures.
        legend_y : float, default=-0.175
            Legend position regarding the plot y-axis applied if ``legend=True``.
        kwargs_scatterplot : dict, optional
            Dictionary with keyword arguments for adjusting scatter plot (:func:`matplotlib.pyplot.scatter`).

        Returns
        -------
        ax : plt.Axes
            PCA plot axes object.

        See Also
        -------
        * :class:`dPULearn` for details on the data structure of ``df_pu``.
        * :func:`matplotlib.pyplot.scatter` for scatter plot arguments.

        Examples
        --------
        .. include:: examples/dpul_plot_pca.rst
        """
        # Check input
        ut.check_df(name="df_pu", df=df_pu, cols_requiered=[ut.COL_SELECTION_VIA], accept_none=True, accept_nan=True)
        n_pc = len([x for x in list(df_pu) if "PC" in x and not "abs_dif" in x])
        if n_pc < 2:
            raise ValueError(f"'df_pu' should contain at least two PCs (n={n_pc}).")
        ut.check_labels(labels=labels) # Pre-check if proper format
        vals_requiered = [0, 1] if 2 not in set(labels) else [0, 1, 2]
        labels = ut.check_labels(labels=labels, vals_requiered=vals_requiered, allow_other_vals=False)
        ut.check_figsize(figsize=figsize, accept_none=True)
        ut.check_number_range(name="pc_x", val=pc_x, min_val=1, max_val=n_pc, just_int=True)
        ut.check_number_range(name="pc_y", val=pc_y, min_val=1, max_val=n_pc, just_int=True)
        ut.check_bool(name="show_pos_mean_x", val=show_pos_mean_x)
        ut.check_bool(name="show_pos_mean_y", val=show_pos_mean_y)
        ut.check_bool(name="legend", val=legend)
        ut.check_number_val(name="legend_y", val=legend_y)
        ut.check_list_colors(name="colors", val=colors, accept_none=True, min_n=2, max_n=3)
        names = ut.check_list_like(name="names", val=names, accept_none=True, check_all_str_or_convertible=True)
        ut.check_dict(name="kwargs_scatterplot", val=kwargs_scatterplot, accept_none=True)
        check_match_df_pu_labels(df_pu=df_pu, labels=labels)
        # Set defaults colors and names
        default_names =  ["Identified negatives", "Positives", "Unlabeled"]
        default_colors = [ut.COLOR_REL_NEG, ut.COLOR_POS, ut.COLOR_UNL]
        names, colors, _args_scatter = _adjust_plot_args(names=names, colors=colors,
                                                         args_scatter=kwargs_scatterplot,
                                                         default_names=default_names,
                                                         default_colors=default_colors)
        # Plotting
        try:
            ax = plot_pca(df_pu=df_pu, labels=labels,
                          figsize=figsize, pc_x=pc_x, pc_y=pc_y,
                          show_pos_mean_x=show_pos_mean_x, show_pos_mean_y=show_pos_mean_y,
                          names=names, colors=colors,
                          legend=legend, legend_y=legend_y, args_scatter=_args_scatter)
        except Exception as e:
            str_error = f"Following error occurred due to plt.scatter() function: {e}"
            raise ValueError(str_error)
        return ax




from ._aamut import AAMut
from ._aamut_plot import AAMutPlot
from ._seqmut import SeqMut
from ._seqmut_plot import SeqMutPlot

__all__ = [
    "AAMut",
    "AAMutPlot",
    "SeqMut",
    "SeqMutPlot",
]


"""
This is a script for the frontend of the plotting utility function to obtain AAanalysis colormaps.
The backend is in general utility module to provide function to remaining AAanalysis modules.
"""
from typing import Union, List, Tuple, Optional
from aaanalysis import utils as ut


# II Main function
def plot_get_cmap(name: str = "CPP",
                  n_colors: int = 101,
                  facecolor_dark: bool = False,
                  ) -> Union[List[Tuple[float, float, float]], List[str]]:
    """
    Get colormaps specified for AAanalysis.

    Parameters
    ----------
    name : {'CPP', 'SHAP'}, default='CPP'
        The name of the AAanalysis color palettes.

         - ``CPP``: Continuous colormap for CPP plots.
         - ``SHAP``: Continuous colormap for CPP-SHP plots.

    n_colors : int, default=101
        Number of colors. Must be at least 3.
    facecolor_dark : bool, optional
        Whether central color in is black (if ``True``) or white (if ``False``).

    Returns
    -------
    cmap
        List with colors given as RGB tuples.

    See Also
    --------
    * `Plotting Prelude <plotting_prelude.html>`_.
    * `Matplotlib color names <https://matplotlib.org/stable/gallery/color/named_colors.html>`_
    * :func:`seaborn.color_palette` function to generate a color palette in seaborn.
    * :func:`seaborn.light_palette function` to generate a lighter color palettes.
    * The `SHAP <shap:mod:shap>`_ package.

    Examples
    --------
    .. include:: examples/plot_get_cmap.rst
    """
    # Check input
    list_names = [ut.STR_CMAP_CPP, ut.STR_CMAP_SHAP]
    if name not in list_names:
        raise ValueError(f"'name' must be one of following: {list_names}")
    ut.check_number_range(name="n_colors", val=n_colors, min_val=3, just_int=True)
    ut.check_bool(name="facecolor_dark", val=facecolor_dark)
    # Get colormaps
    cmap = ut.plot_get_cmap_(cmap=name, n_colors=n_colors, facecolor_dark=facecolor_dark)
    return cmap




"""
This is a script for plot checking utility functions.
"""
import re
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

from ._utils import add_str
import aaanalysis._utils.check_type as check_type
import numpy as np

# Helper functions
def _is_valid_hex_color(val):
    """Check if a value is a valid hex color."""
    return isinstance(val, str) and re.match(r'^#[0-9A-Fa-f]{6}$', val)


def _is_valid_rgb_tuple(val):
    """Check if a value is a valid RGB tuple."""
    return (isinstance(val, tuple) and len(val) == 3 and
            all(isinstance(c, (int, float)) and 0 <= c <= 255 for c in val))


# Check figure
def check_fig(fig=None, accept_none=False, str_add=None):
    """Check if the provided value is a matplotlib Figure instance or None."""
    import matplotlib.figure
    if accept_none and fig is None:
        return None
    if not isinstance(fig, matplotlib.figure.Figure):
        str_error = f"'fig' (type={type(fig)}) should be mpl.figure.Figure or None."
        if str_add:
            str_error += " " + str_add
        raise ValueError(str_error)
    return fig


def check_ax(ax=None, accept_none=False, str_add=None, return_first=False):
    """Check if the provided value is a matplotlib Axes instance or None."""
    import matplotlib.axes
    if accept_none and ax is None:
        return None
    if not isinstance(ax, matplotlib.axes.Axes):
        str_error = add_str(str_error=f"'ax' (type={type(ax)}) should be mpl.axes.Axes or None.",
                            str_add=str_add)
        raise ValueError(str_error)
    if return_first:
        if isinstance(ax, (list, tuple, np.ndarray)) and len(ax) > 0:
            ax = ax[0]
    return ax


def check_figsize(figsize=None, accept_none=False, str_add=None):
    """Check size of figure"""
    if accept_none and figsize is None:
        return None # skip check
    check_type.check_tuple(name="figsize", val=figsize, n=2, str_add=str_add)
    args = dict(min_val=1, just_int=False, str_add=str_add)
    check_type.check_number_range(name="figsize:width", val=figsize[0], **args)
    check_type.check_number_range(name="figsize:height", val=figsize[1], **args)


def check_grid_axis(grid_axis="y", accept_none=True, str_add=None):
    if accept_none and grid_axis is None:
        return None # Skip test
    list_grid_axis = ["y", "x", "both"]
    if grid_axis not in list_grid_axis:
        str_error = add_str(str_error=f"'grid_axis' ({grid_axis}) should be one of following: {list_grid_axis}",
                            str_add=str_add)
        raise ValueError(str_error)


def check_font_weight(name="font_weight", font_weight=None, accept_none=False, str_add=None):
    if accept_none and font_weight is None:
        return None # Skip test
    list_weight = ["normal", "bold"]
    if font_weight not in list_weight:
        str_error = add_str(str_error=f"'{name}' ({font_weight}) should be one of following: {list_weight}",
                            str_add=str_add)
        raise ValueError(str_error)


def check_fontsize_args(**kwargs):
    """Check fontsize parameters"""
    args_fs = {}
    for name, val in kwargs.items():
        check_type.check_number_range(name=name, val=val, min_val=0, accept_none=True, just_int=False)
        args_fs[name] = val
    return args_fs


# Check min and max values
def check_vmin_vmax(vmin=None, vmax=None, str_add=None):
    """Check if vmin and vmax are valid numbers and vmin is less than vmax."""
    args = dict(accept_none=True, just_int=False, str_add=str_add)
    check_type.check_number_val(name="vmin", val=vmin, **args)
    check_type.check_number_val(name="vmax", val=vmax, **args)
    if vmin is not None and vmax is not None and vmin >= vmax:
        str_error = add_str(str_error=f"'vmin' ({vmin}) < 'vmax' ({vmax}) not fulfilled.",
                            str_add=str_add)
        raise ValueError(str_error)


def check_lim(name="xlim", val=None, accept_none=True, str_add=None):
    """Validate that lim parameter ('xlim' or 'ylim') is tuple with two numbers, where the first is less than the second."""
    if val is None:
        if accept_none:
            return None  # Skip check
        else:
            raise ValueError(f"'{name}' should not be None")
    check_type.check_tuple(name=name, val=val, n=2)
    min_val, max_val = val
    args = dict(just_int=False, str_add=str_add)
    check_type.check_number_val(name=f"{name}:min", val=min_val, **args)
    check_type.check_number_val(name=f"{name}:max", val=max_val, **args)
    if min_val >= max_val:
        str_error = add_str(str_error=f"'{name}:min' ({min_val}) should be < '{name}:max' ({max_val}).",
                            str_add=str_add)
        raise ValueError(str_error)


def check_dict_xlims(dict_xlims=None, n_ax=None, str_add=None):
    """Validate the structure and content of dict_xlims to ensure it contains the correct keys and value formats."""
    if n_ax is None:
        # DEV: Developer warning
        raise ValueError("'n_ax' must be specified")
    if dict_xlims is None:
        return
    check_type.check_dict(name="dict_xlims", val=dict_xlims, str_add=str_add)
    wrong_keys = [x for x in list(dict_xlims) if x not in range(n_ax)]
    if len(wrong_keys) > 0:
        str_error = add_str(str_error= f"'dict_xlims' contains invalid keys: {wrong_keys}. "
                                       f"Valid keys are axis indices from 0 to {n_ax - 1}.",
                            str_add=str_add)
        raise ValueError(str_error)
    for key in dict_xlims:
        check_lim(name="xlim", val=dict_xlims[key], str_add=str_add)


# Check colors
def check_color(name=None, val=None, accept_none=False, str_add=None):
    """Check if the provided value is a valid color for matplotlib."""
    if val is None:
        if accept_none:
            return None # Skip test
        else:
            raise ValueError(f"'{name}' should not be None")
    base_colors = list(mcolors.BASE_COLORS.keys())
    tableau_colors = list(mcolors.TABLEAU_COLORS.keys())
    css4_colors = list(mcolors.CSS4_COLORS.keys())
    all_colors = base_colors + tableau_colors + css4_colors
    if _is_valid_hex_color(val) or _is_valid_rgb_tuple(val):
        return
    elif val not in all_colors:
        str_error = add_str(str_error=f"'{name}' ('{val}') is not a valid color. Chose from following: {all_colors}",
                            str_add=str_add)
        raise ValueError(str_error)


def check_list_colors(name=None, val=None, accept_none=False, min_n=None, max_n=None, str_add=None):
    """Check if color list is valid"""
    if accept_none and val is None:
        return None # Skip check
    val = check_type.check_list_like(name=name, val=val, accept_none=accept_none, accept_str=True, str_add=str_add)
    for l in val:
        check_color(name=name, val=l, accept_none=accept_none, str_add=str_add)
    if min_n is not None and len(val) < min_n:
        str_error = add_str(str_error=f"'{name}' should contain at least {min_n} colors",
                            str_add=str_add)
        raise ValueError(str_error)
    if max_n is not None and len(val) > max_n:
        str_error = add_str(str_error=f"'{name}' should contain no more than {max_n} colors",
                            str_add=str_add)
        raise ValueError(str_error)


def check_dict_color(name="dict_color", val=None, accept_none=False, min_n=None, max_n=None, str_add=None):
    """Check if colors in dict_color are valid"""
    if accept_none and val is None:
        return None # Skip check
    check_type.check_dict(name=name, val=val, accept_none=accept_none)
    for key in val:
        check_color(name=name, val=val[key], accept_none=accept_none)
    if min_n is not None and len(val) < min_n:
        str_error = add_str(str_error=f"'{name}' should contain at least {min_n} colors",
                            str_add=str_add)
        raise ValueError(str_error)
    if max_n is not None and len(val) > max_n:
        str_error = add_str(str_error=f"'{name}' should contain no more than {max_n} colors",
                            str_add=str_add)
        raise ValueError(str_error)


def check_cmap(name=None, val=None, accept_none=False, str_add=None):
    """Check if cmap is a valid colormap for matplotlib."""
    valid_cmaps = plt.colormaps()
    if accept_none and val is None:
        pass
    elif val not in valid_cmaps:
        str_error = add_str(str_error=f"'{name}' ('{val}') is not a valid cmap. Chose from following: {valid_cmaps}",
                            str_add=str_add)
        raise ValueError(str_error)


def check_palette(name=None, val=None, accept_none=False, str_add=None):
    """Check if the provided value is a valid color palette."""
    if isinstance(val, str):
        check_cmap(name=name, val=val, accept_none=accept_none, str_add=str_add)
    elif isinstance(val, list):
        for v in val:
            check_color(name=name, val=v, accept_none=accept_none, str_add=str_add)



"""This is the backend for the ShapExplainer().add_sample_mean_dif() method"""
import numpy as np
import pandas as pd
import aaanalysis.utils as ut

def add_sample_mean_dif_(X, labels=None, label_ref=0, df_feat=None, drop=False,
                         sample_positions=None, names=None, group_average=False):
    """ Compute the feature value difference between selected samples and a reference group average,
     and include into feature DataFrame."""
    if drop:
        # Keep original mean_dif
        columns = [x for x in list(df_feat) if ut.COL_MEAN_DIF not in x or x in [ut.COL_MEAN_DIF, ut.COL_ABS_MEAN_DIF]]
        df_feat = df_feat[columns]
    # Compute mean of reference group
    ref_group_mask = np.asarray([l == label_ref for l in labels])
    ref_group_mean = X[ref_group_mask].mean(axis=0)
    # Generate column names based on 'names'
    if not group_average:
        col_names = [f'{ut.COL_MEAN_DIF}_{name}' for name in names]
    else:
        col_names = [f'{ut.COL_MEAN_DIF}_group' if not names else f'{ut.COL_MEAN_DIF}_{names}']
    # Compute the feature value mean differences
    mean_diffs = []
    if group_average:
        group_mean = X[sample_positions].mean(axis=0)
        mean_diffs.append(group_mean - ref_group_mean)
    else:
        for pos in sample_positions:
            sample_mean = X[pos]
            mean_diff = sample_mean - ref_group_mean
            mean_diffs.append(mean_diff)
    # Insert new columns into DataFrame
    df_feat_mean_diff = pd.DataFrame(data=np.array(mean_diffs).T, columns=col_names)
    df_feat = pd.concat([df_feat, df_feat_mean_diff], axis=1)
    return df_feat



"""
This is a script for utility functions for AAclust object and backend.
"""
import numpy as np
from collections import OrderedDict
from sklearn.metrics import pairwise_distances

# II Main Functions
def _cluster_center(X):
    """Compute cluster center (i.e., arithmetical mean over all data points/observations of a cluster)"""
    return X.mean(axis=0)[np.newaxis, :]


def _cluster_medoid(X, metric="correlation"):
    """Obtain cluster medoids (i.e., scale closest to cluster center used as representative scale for a cluster)"""
    center = _cluster_center(X).reshape(1, -1)
    if metric == "correlation":
        # Create new array with cluster center and given array
        center_X = np.concatenate([_cluster_center(X), X], axis=0)
        # Get index for scale with the highest correlation with cluster center
        medoid_index = np.corrcoef(center_X)[0, 1:].argmax()
    else:
        # Calculating pairwise distances from center to all points in X
        distances = pairwise_distances(center, X, metric=metric)
        # Finding the index of the point with the minimum distance to the center
        medoid_index = np.argmin(distances)
    return medoid_index


def _compute_centers(X, labels=None):
    """Obtain cluster centers and their labels"""
    labels_centers = list(OrderedDict.fromkeys(labels))
    list_masks = [[i == label for i in labels] for label in labels_centers]
    centers = np.concatenate([_cluster_center(X[mask]) for mask in list_masks]).round(3)
    labels_centers = np.array(labels_centers)
    return centers, labels_centers


def _compute_medoids(X, labels=None, metric="correlation"):
    """Obtain cluster medoids and their labels"""
    unique_labels = list(OrderedDict.fromkeys(labels))
    list_masks = [[i == label for i in labels] for label in unique_labels]
    # Calculating medoid for each mask using specified metric
    list_ind_max = [_cluster_medoid(X[mask], metric=metric) for mask in list_masks]
    indices = np.array(range(0, len(labels)))
    # Finding global indices of medoids
    medoid_ind = [indices[m][i] for m, i in zip(list_masks, list_ind_max)]
    # Finding labels and data of medoids
    labels_medoids = np.array([labels[i] for i in medoid_ind])
    medoids = np.array([X[i, :] for i in medoid_ind])
    return medoids, labels_medoids, medoid_ind



"""This is a script for the backend of the TreeModel.fit() method."""
import numpy as np
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestClassifier

import aaanalysis.utils as ut

# II Main methods
# 1. Step: Recursive feature elimination (RFE)
def _recursive_feature_elimination(X, labels=None, step=None, n_feat_max=50, n_feat_min=25, n_cv=5, scoring="f1",
                                   random_state=None, verbose=None, i=None, n_rounds=None, is_preselected=None):
    """Perform Recursive Feature Elimination to select the best feature set."""
    rf = RandomForestClassifier(random_state=random_state)
    selected_features = is_preselected.copy()
    n_total = sum(selected_features)
    n_features = n_total
    best_score = 0
    is_selected = selected_features.copy()
    while n_features > n_feat_min:
        if verbose:
            pct_progress = abs(1 - (n_features - n_feat_min) / (n_total - n_feat_min))
            ut.print_progress(i=i+pct_progress, n=n_rounds)
        rf.fit(X[:, selected_features], labels)
        importances = rf.feature_importances_
        if step is None:
            # Remove all features with the minimum importance
            min_importance = np.min(importances)
            features_to_remove = (importances == min_importance)
        else:
            # Remove a fixed number of the least important features (a maximum of n_features -1)
            _step = min(n_features-1, step)
            indices_to_remove = np.argsort(importances)[:_step]
            features_to_remove = np.zeros_like(importances, dtype=bool)
            features_to_remove[indices_to_remove] = True
        selected_features[np.where(selected_features)[0][features_to_remove]] = False
        n_features -= np.sum(features_to_remove)
        # Evaluate the current feature set
        if n_features <= n_feat_max:
            current_score = np.mean(cross_val_score(rf, X[:, selected_features], labels, scoring=scoring, cv=n_cv))
            if current_score > best_score:
                best_score = current_score
                is_selected = selected_features.copy()
    return is_selected


# 2. Step: Computation of feature importance
def _compute_feature_importance(X, labels=None, is_selected=None, list_model_classes=None,
                                list_model_kwargs=None):
    """Compute the average feature importance across multiple tree-based models."""
    selected_indices = np.where(is_selected)[0]
    X_selected = X[:, selected_indices]
    importances = np.zeros((len(list_model_classes), len(is_selected)))
    list_models = []
    for i, model_class in enumerate(list_model_classes):
        model_kwargs = list_model_kwargs[i] if list_model_kwargs is not None else {}
        model = model_class(**model_kwargs)
        model.fit(X_selected, labels)
        list_models.append(model)
        # Update importances only for selected features
        importances[i, selected_indices] = model.feature_importances_
    # Compute the average feature importance across all models
    avg_importance = np.mean(importances, axis=0)
    return avg_importance, list_models


# Complete fit method
@ut.catch_undefined_metric_warning()
def fit_tree_based_models(X=None, labels=None,
                          list_model_classes=None, list_model_kwargs=None, is_preselected=None,
                          n_rounds=None, use_rfe=True, n_cv=5, n_feat_min=10, n_feat_max=50, metric="accuracy", step=1,
                          verbose=True, random_state=None):
    """Fit tree-based models"""
    n_features = X.shape[1]
    feature_importance = np.empty(shape=(n_rounds, n_features))
    is_selected_rounds = np.ones(shape=(n_rounds, n_features), dtype=bool)
    list_models_rounds = []
    if verbose and use_rfe:
        str_n_sets = "1 set" if n_rounds == 1 else f"{n_rounds} sets"
        start_message = (f"Tree Model starts recursive feature elimination "
                         f"to obtain {str_n_sets} of {n_feat_min} to {n_feat_max} features.")
        ut.print_start_progress(start_message=start_message)
    for i in range(n_rounds):
        if is_preselected is not None:
            is_selected = is_preselected.astype(bool)
        else:
            is_selected = np.ones(n_features, dtype=bool)
        # 1. Step: Recursive feature elimination
        if use_rfe:
            args = dict(labels=labels, n_feat_min=n_feat_min, n_feat_max=n_feat_max, n_cv=n_cv, scoring=metric,
                        step=step, random_state=random_state, verbose=verbose, i=i, n_rounds=n_rounds,
                        is_preselected=is_selected)
            is_selected = _recursive_feature_elimination(X, **args)
        is_selected_rounds[i, :] = is_selected
        # 2. Step: Compute feature importance
        avg_importance, list_models = _compute_feature_importance(X, labels=labels, is_selected=is_selected,
                                                                  list_model_classes=list_model_classes,
                                                                  list_model_kwargs=list_model_kwargs)
        feature_importance[i, :] = avg_importance
        list_models_rounds.append(list_models)
    if verbose and use_rfe:
        end_message = "Tree Model finished recursive feature elimination and saves results."
        ut.print_end_progress(end_message=end_message)
    return feature_importance, is_selected_rounds, list_models_rounds



"""
This is a script for the frontend of the NumericalFeature class, a supportive class for the CPP feature engineering,
including scale and feature filtering methods.

"""
import time
import pandas as pd
import numpy as np
from typing import Optional, Literal, Dict, Union, List, Tuple, Type

import aaanalysis.utils as ut

from ._backend.check_feature import check_df_scales

# I Helper Functions
def check_value_type(value_type=None):
    """Check if value type is valid"""
    list_value_type = ["min", "mean", "median", "max"]
    if value_type not in list_value_type:
        raise ValueError(f"'value_type' ('{value_type}') should be on of following: {list_value_type}")

def check_match_df_scales_letter_new(df_scales=None, letter_new=None):
    """Check if new letter not already in df_scales"""
    alphabet = df_scales.index.to_list()
    if letter_new in alphabet:
        raise ValueError(f"Letter '{letter_new}' already exists in alphabet of 'df_scales': {alphabet}")


# TODO add corr filtering


# II Main Functions
class NumericalFeature:
    """
    Utility feature engineering class to process and filter numerical data structures,
    such as amino acid scales or a feature matrix.
    """

    """
    @staticmethod
    def comp_correlation():

    @staticmethod
    def filter_correlation():
    @staticmethod
    def scale_coverage():
    """

    @staticmethod
    def extend_alphabet(df_scales : pd.DataFrame = None,
                        new_letter : str = None,
                        value_type : Literal["min", "mean", "median", "max"] = "mean",
                        ) -> pd.DataFrame:
        """
        Extend amino acid alphabet of ``df_scales`` by new letter.

        This function adds a new row to the DataFrame, representing the new amino acid letter.
        For each scale (column), it computes a specific statistic (min, mean, median, max) based on the
        values of existing amino acids (rows) and assigns this computed value to the new amino acid.

        Parameters
        ----------
        df_scales : pd.DataFrame, shape (n_letters, n_scales)
            DataFrame of scales with letters typically representing amino acids.
        new_letter : str
            The new letter to be added to the alphabet.
        value_type : {'min', 'mean', 'median', 'max'}, default='mean'
            The type of statistic to compute for the new letter (one of 'min', 'mean', 'median', 'max').

        Returns
        -------
        df_scales : pd.DataFrame, shape (n_letters + 1, n_scales)
            DataFrame with the extended alphabet including the new amino acid letter.

        Examples
        --------
        .. include:: examples/nf_extend_alphabet.rst
        """
        # Check input
        df_scales = df_scales.copy()
        check_df_scales(df_scales=df_scales)
        ut.check_str(name="letter_new", val=new_letter)
        check_value_type(value_type=value_type)
        check_match_df_scales_letter_new(df_scales=df_scales, letter_new=new_letter)
        # Compute the statistic for each scale
        if value_type == "min":
            new_values = df_scales.min()
        elif value_type == "mean":
            new_values = df_scales.mean()
        elif value_type == "median":
            new_values = df_scales.median()
        else:
            new_values = df_scales.max()
        # Add the new letter to the DataFrame
        df_scales.loc[new_letter] = new_values
        return df_scales

    """
    @staticmethod
    def merge_alphabet(df_scales : pd.DataFrame = None,
                       letters_to_merge : List[str] = None,
                       letter_new : str = None,
                       value_type: Literal["min", "mean", "median", "max"] = "mean"
                       ) -> pd.DataFrame:
    """



"""
This a script for general decorators used in AAanalysis.
# Dev: use runtime decorator only for internal methods since they destroy the signature for some IDEs
"""
import warnings
import traceback
from sklearn.exceptions import ConvergenceWarning, UndefinedMetricWarning
import functools
import re

# Helper functions


# Catch Runtime
class CatchRuntimeWarnings:
    """Context manager to catch RuntimeWarnings and store them in a list."""
    def __enter__(self):
        self._warn_list = []
        self._other_warnings = []
        self._showwarning_orig = warnings.showwarning
        warnings.showwarning = self._catch_warning
        return self

    def __exit__(self, exc_type, exc_value, tb):
        warnings.showwarning = self._showwarning_orig
        # Re-issue any other warnings that were caught but not RuntimeWarning
        for warn_message, warn_category, filename, lineno in self._other_warnings:
            warnings.warn_explicit(warn_message, warn_category, filename, lineno)

    def _catch_warning(self, message, category, filename, lineno, file=None, line=None):
        if category == RuntimeWarning:
            line_content = traceback.format_list([(filename, lineno, "", line)])[0].strip()
            warning_msg = f"{message}: {line_content.split(', in')[1]}"
            self._warn_list.append(warning_msg)
        else:
            # Store other warnings for re-issuing later
            self._other_warnings.append((message, category, filename, lineno))

    def get_warnings(self):
        return self._warn_list


def catch_runtime_warnings():
    """Decorator to catch RuntimeWarnings and store them in a list.

    Returns:
        decorated_func: The decorated function
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with CatchRuntimeWarnings() as crw:
                result = func(*args, **kwargs)
            if crw.get_warnings():
                list_warnings = crw.get_warnings()
                n = len(list_warnings)
                summary_msg = f"The following {n} 'RuntimeWarnings' were caught:\n" + "\nRuntimeWarning: ".join(crw.get_warnings())
                warnings.warn(summary_msg, RuntimeWarning)
            return result
        return wrapper

    return decorator


# Catch convergence
class ClusteringConvergenceException(Exception):
    def __init__(self, message, distinct_clusters):
        super().__init__(message)
        self.distinct_clusters = distinct_clusters


def catch_convergence_warning():
    """Decorator to catch ConvergenceWarnings and raise custom exceptions.

    Returns:
        decorated_func: The decorated function.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with warnings.catch_warnings(record=True) as w:
                # Trigger the "always" behavior for ConvergenceWarning
                warnings.simplefilter("always", ConvergenceWarning)
                result = func(*args, **kwargs)  # Call the original function

                # Check if the warning is the one we're interested in
                for warn in w:
                    if issubclass(warn.category, ConvergenceWarning):
                        message = str(warn.message)
                        if "Number of distinct clusters" in message:
                            distinct_clusters = int(message.split("(")[1].split(")")[0].split()[0])
                            raise ClusteringConvergenceException(f"Process stopped due to ConvergenceWarning.", distinct_clusters)
            return result
        return wrapper

    return decorator


# Catch invalid division (could be added to AAclust().comp_medoids())
class InvalidDivisionException(Exception):
    pass


def catch_invalid_divide_warning():
    """Decorator to catch specific RuntimeWarnings related to invalid division
       and raise custom exceptions.

    Returns:
        decorated_func: The decorated function.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with CatchRuntimeWarnings() as crw:
                result = func(*args, **kwargs)
            if crw.get_warnings():
                raise InvalidDivisionException(f"\nError due to 'RuntimeWarning': {crw.get_warnings()[0]}")
            return result
        return wrapper
    return decorator

# Catch UndefinedMetricWarnings
class CatchUndefinedMetricWarning:
    """Context manager to catch and aggregate UndefinedMetricWarnings."""
    def __enter__(self):
        self._warn_set = set()
        self._other_warnings = []
        self._showwarning_orig = warnings.showwarning
        warnings.showwarning = self._catch_warning
        return self

    def __exit__(self, exc_type, exc_value, tb):
        warnings.showwarning = self._showwarning_orig
        for warn_message, warn_category, filename, lineno in self._other_warnings:
            warnings.warn_explicit(warn_message, warn_category, filename, lineno)

    def _catch_warning(self, message, category, filename, lineno, file=None, line=None):
        if category == UndefinedMetricWarning:
            self._warn_set.add(str(message))  # Add message to set (duplicates are automatically handled)
        else:
            self._other_warnings.append((message, category, filename, lineno))

    def get_warnings(self):
        return list(self._warn_set)

def catch_undefined_metric_warning():
    """Decorator to catch and report UndefinedMetricWarnings once per unique message."""
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with CatchUndefinedMetricWarning() as cumw:
                result = func(*args, **kwargs)
            if cumw.get_warnings():
                summary_msg = "The following 'UndefinedMetricWarning' was caught:\n" + "\n".join(cumw.get_warnings())
                summary_msg += ("\n This warning was likely triggered due to 'precision' or 'f1' metrics and "
                                "an imbalanced and/or small dataset.")
                warnings.warn(summary_msg, UndefinedMetricWarning)
            return result
        return wrapper
    return decorator


"""
This is a script for common check function for the explainable AI models.
"""
import numpy as np
import aaanalysis.utils as ut


def check_match_list_model_classes_kwargs(list_model_classes=None, list_model_kwargs=None):
    """Check length match of list_model_classes and list_model_kwargs"""
    n_models = len(list_model_classes)
    n_args = len(list_model_kwargs)
    if n_models != n_args:
        raise ValueError(f"Length of 'list_model_kwargs' (n={n_args}) should match to 'list_model_classes' (n{n_models}")


def check_match_labels_X(labels=None, X=None):
    """Check if labels binary classification task labels"""
    n_samples = X.shape[0]
    # Accept float if fuzzy_labeling is True
    str_add = "Consider setting 'fuzzy_labeling=True'."
    labels = ut.check_labels(labels=labels, len_requiered=n_samples, str_add=str_add)
    unique_labels = set(labels)
    if len(unique_labels) != 2:
        raise ValueError(f"'labels' should contain 2 unique labels ({unique_labels})")
    # The higher label is considered as the positive (test) class
    label_test = list(sorted(unique_labels))[1]
    labels = np.asarray([1 if x == label_test else 0 for x in labels])
    return labels


def check_match_X_is_selected(X=None, is_selected=None):
    """Check if length of X and feature selection mask (is_selected) matches"""
    n_features = X.shape[1]
    n_feat_is_selected = len(is_selected[0])
    if n_features != n_feat_is_selected:
        raise ValueError(f"Number of features from 'X' ({n_features}) does not match "
                         f"with 'is_selected' attribute ({n_feat_is_selected})")



"""
This is a script for the frontend of the SequenceFeature class, a supportive class for the CPP feature engineering.
"""
from typing import Optional, Union, List
import warnings
import pandas as pd

import aaanalysis.utils as ut
from ._backend.check_feature import (check_split_kws,
                                     check_parts_len,
                                     check_match_features_seq_parts,
                                     check_match_df_seq_jmd_len,
                                     check_df_parts,
                                     check_match_df_parts_features,
                                     check_match_df_parts_df_scales,
                                     check_df_scales,
                                     check_match_df_scales_features,
                                     check_match_df_scales_df_cat,
                                     check_df_cat,
                                     check_match_df_cat_features)
from ._backend.cpp.utils_feature import (get_df_parts_, remove_entries_with_gaps_,
                                         get_positions_, get_amino_acids_,
                                         get_feature_matrix_, get_df_pos_, get_df_pos_parts_)
from ._backend.cpp.sequence_feature import (get_split_kws_, get_features_, get_feature_names_, get_df_feat_)


# I Helper Functions
def check_split_types(split_types=None):
    """Check if split types valid (Segment, Pattern, or PeriodicPattern)"""
    split_types = ut.check_list_like(name="split_types", val=split_types, accept_str=True, accept_none=True)
    if split_types is None:
        split_types = ut.LIST_SPLIT_TYPES
    wrong_split_type = [x for x in split_types if x not in ut.LIST_SPLIT_TYPES]
    if len(wrong_split_type) > 0:
        raise ValueError(f"Wrong 'split_types' ({wrong_split_type}). Chose from {ut.LIST_SPLIT_TYPES}")
    return split_types


def check_steps(steps=None, steps_name="steps_pattern", len_min=2, fixed_len=False):
    """Sort steps and warn if empty list"""
    if steps is None:
        return steps # Skip tests
    steps = list(sorted(steps))
    if len(steps) < len_min:
        if fixed_len:
            raise ValueError(f"'{steps_name}' ({steps}) should contain exactly {len_min} non-negative integers.")
        else:
            raise ValueError(f"'{steps_name}' ({steps}) should contain >= {len_min} non-negative integers.")
    return steps


def warn_creation_of_feature_matrix(features=None, df_parts=None, name="Feature matrix"):
    """Warn if feature matrix gets too large"""
    n_feat = len(features)
    n_samples = len(df_parts)
    n_vals = n_feat * n_samples
    ut.print_out(f"'{name}' for {n_feat} features and {n_samples} samples will be created.")
    if n_vals > 1000 * 1000:
        warning = f"Feature matrix with n={n_vals}>=10^6 values will be created, which will take some time.\n" \
                  "It is recommended to create a feature matrix for a pre-selected number features " \
                  "so that 10^6 values are not exceeded."
        warnings.warn(warning)


def check_match_labels_label_test_label_ref(labels=None, label_test=1, label_ref=0):
    """Check if labels only contains label_test and label_ref"""
    wrong_labels = [x for x in labels if x not in [label_ref, label_test]]
    unique_wrong_labels = list(set(wrong_labels))
    n_wrong_labels = len(unique_wrong_labels)
    if n_wrong_labels > 0:
        raise ValueError(f"'labels' contains {n_wrong_labels} wrong labels: {unique_wrong_labels}")


def check_match_df_parts_label_test_label_ref(df_parts=None, labels=None, label_test=1, label_ref=0):
    """Check if 'jmd_n', 'tmd', and 'jmd_c' in df_parts if amino acid for label_test or label_ref should be retrieved"""
    list_parts = list(df_parts)
    requiered_parts = ["jmd_n", "tmd", "jmd_c"]
    mask_test = [x == label_test for x in labels]
    mask_ref = [x == label_ref for x in labels]
    if sum(mask_test) == 1:
        missing_parts = [x for x in requiered_parts if x not in list_parts]
        if len(missing_parts) > 0:
            raise ValueError(f"'df_parts' misses '{missing_parts}' parts necessary to retrieve amino acid positions"
                             f" for 'label_test' ({label_test}) if only one sample of it occurs in 'labels'."
                             f"\n Add them to the current parts of 'df_parts': {list_parts}")
    if sum(mask_ref) == 1:
        missing_parts = [x for x in requiered_parts if x not in list_parts]
        if len(missing_parts) > 0:
            raise ValueError(f"'df_parts' misses '{missing_parts}' parts necessary to retrieve amino acid positions"
                             f" for 'label_ref' ({label_ref}) if only one sample of it occurs in 'labels'."
                             f"\n Add them to the current parts of 'df_parts': {list_parts}")


def check_col_cat(col_cat=None):
    """Check if col_cat valid column from df_feat"""
    if col_cat not in ut.COLS_FEAT_SCALES:
        raise ValueError(f"'col_cat' {col_cat} should be one of the following: {ut.COLS_FEAT_SCALES}")


def check_col_val(col_val=None):
    """Check if col_val valid column from df_feat"""
    cols_feat = ut.COLS_FEAT_STAT + ut.COLS_FEAT_WEIGHT
    if col_val not in cols_feat:
        raise ValueError(f"'col_val' {col_val} should be one of the following: {cols_feat}")


# II Main Functions
class SequenceFeature:
    """
    Utility feature engineering class using sequences to create :class:`CPP` feature components (**Parts**, **Splits**,
    and  **Scales**) and data structures [Breimann24c]_.

    The three feature components are the primary input for the :class:`aaanalysis.CPP` class and define
    Comparative Physicochemical Profiling (CPP) features.

    Notes
    -----
    Feature Components:
        - **Part**: A continuous subset of a sequence, such as a protein domain.
        - **Split**: Continuous or discontinuous subset of a **Part**, either segment or pattern.
        - **Scale**: A physicochemical scale, i.e., a set of numerical values (typically [0-1]) assigned to amino acids.

    Main Parts:
        We define three main parts from which each other part can be derived from:

        - **TMD (target middle domain)**: Protein domain of interest with varying length.
        - **JMD-N (juxta middle domain N-terminal)**: Protein domain or sequence region directly
          N-terminally next to the TMD, typically set to a fixed length (10 by default).
        - **JMD-C (juxta middle domain C-terminal)**: Protein domain or sequence region directly
          C-terminally next to the TMD, typically set to a fixed length (10 by default).

    Feature: Part + Split + Scale
        Physicochemical property (expressed as numerical scale) present at distinct amino acid
        positions within a protein sequence. The positions are obtained by splitting sequence parts
        into segments or patterns.

    Feature value: Realization of a Feature
        For a given sequence, a feature value is the average of a physicochemical scale over
        all amino acids obtained by splitting a sequence part.

    Valid sequence parts:
        - ``tmd``: Target Middle Domain (TMD).
        - ``tmd_e``: TMD extended N- and C-terminally by a number of residues, defined by the ``ext_len`` configuration option.
        - ``tmd_n``: N-terminal half of the TMD.
        - ``tmd_c``: C-terminal half of the TMD.
        - ``jmd_n``: N-terminal Juxt Middle Domain (JMD).
        - ``jmd_c``: C-terminal JMD.
        - ``ext_c``: Extended C-terminal region.
        - ``ext_n``: Extended N-terminal region.
        - ``tmd_jmd``: Combination of JMD-N, TMD, and JMD-C.
        - ``jmd_n_tmd_n``: Combination of JMD-N and N-terminal half of TMD.
        - ``tmd_c_jmd_c``: Combination of C-terminal half of TMD and JMD-C.
        - ``ext_n_tmd_n``: Extended N-terminal region and N-terminal half of TMD.
        - ``tmd_c_ext_c``: C-terminal half of TMD and extended C-terminal region.

    Default parts:
        The following three parts are provided by default: ``tmd``, ``jmd_n_tmd_n``, ``tmd_c_jmd_c``.

    """

    def __init__(self,
                 verbose: bool = True
                 ):
        """
        Parameters
        ----------
        verbose : bool, default=True
            If ``True``, verbose outputs are enabled.
        """
        self.verbose = ut.check_verbose(verbose)

    # Part and Split methods
    def get_df_parts(self,
                     df_seq: pd.DataFrame = None,
                     list_parts: Optional[Union[str, List[str]]] = None,
                     all_parts: bool = False,
                     jmd_n_len: Union[int, None] = 10,
                     jmd_c_len: Union[int, None] = 10,
                     remove_entries_with_gaps: bool = False
                     ) -> pd.DataFrame:
        """
        Create DataFrane with selected sequence parts.

        Parameters
        ----------
        df_seq : pd.DataFrame, shape (n_samples, n_seq_info)
            DataFrame containing an ``entry`` column with unique protein identifiers and sequence information
            in a distinct ``Position-based``, ``Part-based``, ``Sequence-based``, or ``Sequence-TMD-based`` format.
        list_parts: list of str, default={``tmd``, ``jmd_n_tmd_n``, ``tmd_c_jmd_c``}
            Names of sequence parts that should be obtained for sequences from ``df_seq``.
        jmd_n_len: int, default=10
            Length of JMD-N in number of amino acids. If ``None``, ``jmd_n`` and ``jmd_c`` should be given.
        jmd_c_len: int, default=10
            Length of JMD-N in number of amino acids. If ``None``, ``jmd_n`` and ``jmd_c`` should be given.
        all_parts: bool, default=False
            Whether to create DataFrame with all possible sequence parts (if ``True``) or parts given by ``list_parts``.
        remove_entries_with_gaps: bool, default=False
            Whether to exclude entries containing missing residues in their sequence parts (if ``True``),
            usually resulting from sequences being too short.

        Returns
        -------
        df_parts: pd.DataFrame, shape (n_samples, n_parts)
            Sequence parts DataFrame.

        See Also
        --------
        * :class:`aaanalysis.SequenceFeature` for definition of parts, and lists of all existing and default parts.

        Notes
        -----
        * If ``ext_len`` in aaanalysis.options is not set to > 0, following parts containing extended tmd are not
          considered for ``all_parts=True``: ['tmd_e', 'ext_c', 'ext_n', 'ext_n_tmd_n', 'tmd_c_ext_c'].
        * ``jmd_n_len`` and ``jmd_c_len`` must be both given, except for the part-based format.

        Formats for ``df_seq`` are differentiated by their respective columns:

        **Position-based format**
            - 'sequence': The complete amino acid sequence.
            - 'tmd_start': Starting positions of the TMD in the sequence.
            - 'tmd_stop': Ending positions of the TMD in the sequence.

        **Part-based format**
            - 'jmd_n': Amino acid sequence for JMD-N.
            - 'tmd': Amino acid sequence for TMD.
            - 'jmd_c': Amino acid sequence for JMD-C.

        **Sequence-TMD-based format**
            - 'sequence' and 'tmd' columns.

        **Sequence-based format**
            - Only the 'sequence' column.

        Examples
        --------
        .. include:: examples/sf_get_df_parts.rst
        """
        # Check input
        check_parts_len(jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len, accept_none_tmd_len=True)
        ut.check_df_seq(df_seq=df_seq)
        ut.check_bool(name="all_parts", val=all_parts)
        list_parts = ut.check_list_parts(list_parts=list_parts, all_parts=all_parts, accept_none=True)
        df_seq = check_match_df_seq_jmd_len(df_seq=df_seq, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        # Create df parts
        df_parts = get_df_parts_(df_seq=df_seq, list_parts=list_parts, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        if remove_entries_with_gaps:
            n_before = len(df_parts)
            df_parts = remove_entries_with_gaps_(df_parts=df_parts)
            n_removed = n_before - len(df_parts)
            if n_removed > 0 and self.verbose:
                warnings.warn(f"{n_removed} entries have been removed from 'df_seq' due to introduced gaps.")
        if len(df_parts) == 0:
            raise ValueError(f"All entries have been removed from 'df_seq'. "
                             f"Reduce 'jmd_n_len' ({jmd_n_len}) and 'jmd_c_len' ({jmd_c_len}) settings.")
        return df_parts

    @staticmethod
    def get_split_kws(split_types: Union[str, List[str]] = None,
                      n_split_min: int = 1,
                      n_split_max: int = 15,
                      steps_pattern: Optional[List[int]] = None,
                      n_min: int = 2,
                      n_max: int = 4,
                      len_max: int = 15,
                      steps_periodicpattern: Optional[List[int]] = None,
                      ) -> dict:
        """
        Create dictionary with kwargs for three split types:

            - **Segment**: continuous sub-sequence.
            - **Pattern**: non-periodic discontinuous sub-sequence
            - **PeriodicPattern**: periodic discontinuous sub-sequence.

        Parameters
        ----------
        split_types: list of str, default=[``Segment``, ``Pattern``, ``PeriodicPattern``]
            Split types for which parameter dictionary should be generated.
        n_split_min: int, default=1
            Number to specify the greatest ``Segment``. Should be > 0.
        n_split_max: int, default=15,
            Number to specify the smallest ``Segment``. Should be > ``n_split_min``.
        steps_pattern: list of int, default=[3, 4], optional
            Possible steps sizes for ``Pattern``. Should contain at least 1 non-negative integers
            if ``Pattern`` split_type is used. If ``None``, default is used.
        n_min: int, default=2
            Minimum number of steps for ``Pattern``. Should be <= ``n_max``.
        n_max: int, default=4
            Maximum number of steps for ``Pattern``. Should be >= ``n_min``.
        len_max: int, default=10
            Maximum length in amino acid position for ``Pattern`` by varying start position.
            Should be > min(``steps_pattern``).
        steps_periodicpattern: list of int, default=[3, 4], optional
            Size of odd and even steps for ``PeriodicPattern``. Should contain two non-negative integers if
            ``PeriodicPattern`` split_type is used. If ``None``, default is used.

        Returns
        -------
        split_kws: dict
            Nested dictionary with parameters for chosen split_types:

            - Segment: {n_split_min:1, n_split_max=15}
            - Pattern: {steps=[3, 4], n_min=2, n_max=4, len_max=15}
            - PeriodicPattern: {steps=[3, 4]}

        Examples
        --------
        .. include:: examples/sf_get_split_kws.rst
        """
        # Check input
        split_types = check_split_types(split_types=split_types)
        args_int = dict(n_split_min=n_split_min, n_split_max=n_split_max, n_min=n_min, n_max=n_max, len_max=len_max)
        for name in args_int:
            ut.check_number_range(name=name, val=args_int[name], just_int=False, min_val=1)
        steps_pattern = ut.check_list_like(name="steps_pattern", val=steps_pattern,
                                           accept_none=True, check_all_non_neg_int=True)
        steps_periodicpattern = ut.check_list_like(name="steps_periodicpattern", val=steps_periodicpattern,
                                                   accept_none=True, check_all_non_neg_int=True)
        steps_pattern = check_steps(steps=steps_pattern, steps_name="steps_pattern", len_min=1, fixed_len=False)
        steps_periodicpattern = check_steps(steps=steps_periodicpattern, steps_name="steps_periodicpattern",
                                            len_min=2, fixed_len=True)
        # Create kws for splits
        split_kws = get_split_kws_(n_split_min=n_split_min,
                                   n_split_max=n_split_max,
                                   steps_pattern=steps_pattern,
                                   n_min=n_min,
                                   n_max=n_max,
                                   len_max=len_max,
                                   steps_periodicpattern=steps_periodicpattern,
                                   split_types=split_types)
        # Post check
        check_split_kws(split_kws=split_kws)
        return split_kws

    # Feature methods
    def get_df_feat(self,
                    features: ut.ArrayLike1D = None,
                    df_parts: pd.DataFrame = None,
                    labels: ut.ArrayLike1D = None,
                    label_test : int = 1,
                    label_ref : int = 0,
                    df_scales: Optional[pd.DataFrame] = None,
                    df_cat: Optional[pd.DataFrame] = None,
                    start: int = 1,
                    tmd_len: int = 20,
                    jmd_c_len: int = 10,
                    jmd_n_len: int = 10,
                    accept_gaps: bool = False,
                    parametric: bool = False,
                    n_jobs: Union[int, None] = 1,
                    ) -> pd.DataFrame:
        """
        Create feature DataFrame for given features.

        Depending on the provided labels, the DataFrame is created for one of the three following cases:

            1. Group vs group comparison
            2. Sample vs group comparison
            3. Sample vs sample comparison

        * For the group vs group comparison, the general feature position will be provided.
        * For sample vs group or sample vs sample comparison, the amino acid segments
          and patterns for the respective sample from the test dataset (label = 1) will be given.

        Parameters
        ----------
        features : array-like, shape (n_features,)
            Ids of features for which ``df_feat`` should be created.
        df_parts : pd.DataFrame, shape (n_samples, n_parts)
            DataFrame with sequence parts. Must cover all parts in ``features``.
        labels: array-like, shape (n_samples,)
            Class labels for samples in ``df_parts``. Should contain only two different integer label values,
            representing test and reference group (typically, 1 and 0).
        label_test : int, default=1,
            Class label of test group in ``labels``.
        label_ref : int, default=0,
            Class label of reference group in ``labels``.
        df_scales : pd.DataFrame, shape (n_letters, n_scales), optional
            DataFrame of scales with letters typically representing amino acids. Default from :meth:`load_scales`
            unless specified in ``options['df_scales']``.
        df_cat : pd.DataFrame, shape (n_scales, n_scales_info), optional
            DataFrame of categories for physicochemical scales. Must contain all scales from ``df_scales``.
            Default from :meth:`load_scales` with ``name='scales_cat'``, unless specified in ``options['df_cat']``.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD (>0).
        jmd_n_len : int, default=10
            Length of JMD-N (>=0).
        jmd_c_len : int, default=10
            Length of JMD-C (>=0).
        parametric : bool, default=False
            Whether to use parametric (T-test) or non-parametric (Mann-Whitney-U-test) test for p-value computation.
        accept_gaps : bool, default=False
            Whether to accept missing values by enabling omitting for computations (if True).
        n_jobs : int, None, or -1, default=1
            Number of CPU cores used for multiprocessing. If ``None``, the number is optimized automatically.
            If ``-1``, the number is set to all available cores.

        Returns
        -------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.

        Notes
        -----
        * Use parallel processing only for high number of features (>~1000 features per core)
        * For sample vs group or sample vs sample comparison, ``df_parts`` must comprise ``jmd_n``, ``tmd``, and
          ``jmd_c`` sequence parts as well as all parts in features.

        See Also
        --------
        * The :meth:`CPP.run` method for creating and filtering CPP features for discriminating between
          two groups of sequences.

        Examples
        --------
        .. include:: examples/sf_get_df_feat.rst
        """
        # Load defaults
        if df_scales is None:
            df_scales = ut.load_default_scales()
        if df_cat is None:
            df_cat = ut.load_default_scales(scale_cat=True)
        # Check input
        check_df_parts(df_parts=df_parts)
        check_df_scales(df_scales=df_scales)
        features = ut.check_features(features=features, list_parts=list(df_parts), list_scales=list(df_scales))
        ut.check_number_val(name="label_test", val=label_test, just_int=True)
        ut.check_number_val(name="label_ref", val=label_ref, just_int=True)
        labels = ut.check_labels(labels=labels, vals_requiered=[label_test, label_ref],
                                 len_requiered=len(df_parts), allow_other_vals=False)
        ut.check_number_val(name="start", val=start, just_int=True, accept_none=False)
        args_len, _ = check_parts_len(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        ut.check_bool(name="parametric", val=parametric)
        ut.check_bool(name="accept_gaps", val=accept_gaps)
        check_match_df_parts_features(df_parts=df_parts, features=features)
        check_match_df_scales_features(df_scales=df_scales, features=features)
        check_match_features_seq_parts(features=features, **args_len)
        check_match_df_scales_df_cat(df_scales=df_scales, df_cat=df_cat, verbose=self.verbose)
        df_parts = check_match_df_parts_df_scales(df_scales=df_scales, df_parts=df_parts, accept_gaps=accept_gaps)
        check_match_labels_label_test_label_ref(labels=labels, label_test=label_test, label_ref=label_ref)
        check_match_df_parts_label_test_label_ref(df_parts=df_parts, labels=labels,
                                                  label_test=label_test, label_ref=label_ref)
        n_jobs = ut.check_n_jobs(n_jobs=n_jobs)
        # User warning
        if self.verbose:
            warn_creation_of_feature_matrix(features=features, df_parts=df_parts, name="df_feat")
        # Get sample difference to reference group
        df_feat = get_df_feat_(features=features, df_parts=df_parts, labels=labels,
                               label_test=label_test, label_ref=label_ref,
                               df_scales=df_scales, df_cat=df_cat,
                               accept_gaps=accept_gaps, parametric=parametric,
                               start=start, jmd_n_len=jmd_n_len, tmd_len=tmd_len, jmd_c_len=jmd_c_len,
                               n_jobs=n_jobs)
        return df_feat

    def feature_matrix(self,
                       features: ut.ArrayLike1D = None,
                       df_parts: pd.DataFrame = None,
                       df_scales: Optional[pd.DataFrame] = None,
                       accept_gaps: bool = False,
                       n_jobs: Union[int, None] = 1,
                       ) -> ut.ArrayLike2D:
        """
        Create feature matrix for given feature ids and sequence parts.

        Parameters
        ----------
        features : array-like, shape (n_features,)
            Ids of features for which matrix of feature values should be created.
        df_parts : pd.DataFrame, shape (n_samples, n_parts)
            DataFrame with sequence parts.
        df_scales : pd.DataFrame, shape (n_letters, n_scales), optional
            DataFrame of scales with letters typically representing amino acids. Default from :meth:`load_scales`
            unless specified in ``options['df_scales']``.
        accept_gaps: bool, default=False
            Whether to accept missing values by enabling omitting for computations (if ``True``).
        n_jobs : int, None, or -1, default=1
            Number of CPU cores used for multiprocessing. If ``None``, the number is optimized automatically.
            If ``-1``, the number is set to all available cores.

        Returns
        -------
        feat_matrix: array-like , shape (n_samples, n_features)
            Feature values for samples.

        Notes
        -----
        * Use parallel processing only for high number of features (>~1000 features per core)

        Examples
        --------
        .. include:: examples/sf_feature_matrix.rst
        """
        # Load defaults
        if df_scales is None:
            df_scales = ut.load_default_scales()
        # Check input
        check_df_scales(df_scales=df_scales)
        check_df_parts(df_parts=df_parts)
        features = ut.check_features(features=features, list_parts=list(df_parts), list_scales=list(df_scales))
        ut.check_bool(name="accept_gaps", val=accept_gaps)
        n_jobs = ut.check_n_jobs(n_jobs=n_jobs)
        check_match_df_parts_features(df_parts=df_parts, features=features)
        check_match_df_scales_features(df_scales=df_scales, features=features)
        df_parts = check_match_df_parts_df_scales(df_scales=df_scales, df_parts=df_parts, accept_gaps=accept_gaps)
        # User warning
        if self.verbose:
            warn_creation_of_feature_matrix(features=features, df_parts=df_parts)
        # Create feature matrix using parallel processing
        feat_matrix = get_feature_matrix_(features=features,
                                          df_parts=df_parts,
                                          df_scales=df_scales,
                                          accept_gaps=accept_gaps,
                                          n_jobs=n_jobs)
        return feat_matrix

    def get_features(self,
                     list_parts: Optional[List[str]] = None,
                     all_parts: bool = False,
                     split_kws: Optional[dict] = None,
                     list_scales: Optional[List[str]] = None,
                     ) -> List[str]:
        """
        Create list of all feature ids for given Parts, Splits, and Scales.

        Parameters
        ----------
        list_parts: list of str, default=["tmd", "jmd_n_tmd_n", "tmd_c_jmd_c"]
            Names of sequence parts which should be created (e.g., 'tmd'). Length should be >= 1.
        all_parts: bool, default=False
            Whether to create DataFrame with all possible sequence parts (if ``True``) or parts given by ``list_parts``.
        split_kws : dict, optional
            Dictionary with parameter dictionary for each chosen split_type. Default from :meth:`SequenceFeature.get_split_kws`.
        list_scales : list of str, optional
            Names of scales. Default scales from :meth:`load_scales` with ``name='scales'``.

        Returns
        -------
        features: list of str
            Ids of all possible features for combination of Parts, Splits, and Scales with form: PART-SPLIT-SCALE

        Notes
        -----
        * If ``ext_len`` in aaanalysis.options is not set to > 0, following parts containing extended tmd are not
          considered for ``all_parts=True``: ['tmd_e', 'ext_c', 'ext_n', 'ext_n_tmd_n', 'tmd_c_ext_c'].

        Examples
        --------
        .. include:: examples/sf_get_features.rst
        """
        # Load defaults
        if list_scales is None:
            list_scales = list(ut.load_default_scales())
        if split_kws is None:
            split_kws = self.get_split_kws()
        # Check input
        ut.check_bool(name="all_parts", val=all_parts)
        list_parts = ut.check_list_parts(list_parts=list_parts, all_parts=all_parts)
        check_split_kws(split_kws=split_kws)
        list_scales = ut.check_list_like(name="list_scales", val=list_scales, accept_none=False)
        # Get features
        features = get_features_(list_parts=list_parts, split_kws=split_kws, list_scales=list_scales)
        return features

    @staticmethod
    def get_feature_names(features: ut.ArrayLike1D = None,
                          df_cat: Optional[pd.DataFrame] = None,
                          start: int = 1,
                          tmd_len: int = 20,
                          jmd_n_len: int = 10,
                          jmd_c_len: int = 10,
                          ) -> List[str]:
        """
        Convert feature ids (PART-SPLIT-SCALE) into feature names (scale name [positions]).

        Parameters
        ----------
        features : array-like, shape (n_features,)
            List of feature ids (>0).
        df_cat : pd.DataFrame, shape (n_scales, n_scales_info), optional
            DataFrame of categories for physicochemical scales. Must contain all scales from ``df_scales``.
            Default from :meth:`load_scales` with ``name='scales_cat'``, unless specified in ``options['df_cat']``.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD (>0).
        jmd_n_len : int, default=10
            Length of JMD-N (>=0).
        jmd_c_len : int, default=10
            Length of JMD-C (>=0).

        Returns
        -------
        feat_names: list of str
            Names of features.

        Notes
        -----
        * Length parameters (``tmd_len``, ``jmd_n_len``, ``jmd_c_len``) must match with ids in ``features``.
        * Positions are given depending on the three split types:

            - Segment: [first...last]
            - Pattern: [all positions]
            - PeriodicPattern: [first..step1/step2..last]

        Examples
        --------
        .. include:: examples/sf_get_feature_names.rst
        """
        # Load defaults
        if df_cat is None:
            df_cat = ut.load_default_scales(scale_cat=True)
        # Check input
        features = ut.check_features(features=features)
        check_df_cat(df_cat=df_cat)
        ut.check_number_val(name="start", val=start, just_int=True, accept_none=False)
        args_len, _ = check_parts_len(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        check_match_df_cat_features(df_cat=df_cat, features=features)
        check_match_features_seq_parts(features=features, **args_len)
        # Get feature names
        feat_names = get_feature_names_(features=features,
                                        df_cat=df_cat,
                                        start=start,
                                        tmd_len=tmd_len,
                                        jmd_c_len=jmd_c_len,
                                        jmd_n_len=jmd_n_len)
        return feat_names

    @staticmethod
    def get_feature_positions(features: ut.ArrayLike1D = None,
                              start: int = 1,
                              tmd_len: int = 20,
                              jmd_n_len: int = 10,
                              jmd_c_len: int = 10,
                              tmd_seq: Optional[str] = None,
                              jmd_n_seq: Optional[str] = None,
                              jmd_c_seq: Optional[str] = None,
                              ) -> ut.ArrayLike1D:
        """
        Create for features a list of corresponding positions or amino acids.

        Parameters
        ----------
        features : array-like, shape (n_features,)
            List of feature ids.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD (>0).
        jmd_n_len : int, default=10
            Length of JMD-N (>=0).
        jmd_c_len : int, default=10
            Length of JMD-C (>=0).
        tmd_seq : str, optional
            Sequence of TMD. If given, respective amino acid segments/patterns will be returned instead of positions.
        jmd_n_seq : str, optional
            Sequence of JMD-N. If given, respective amino acid segments/patterns will be returned instead of positions.
        jmd_c_seq : str, optional
            Sequence of JMD-C. If given, respective amino acid segments/patterns will be returned instead of positions.

        Returns
        -------
        list_pos or list_aa
            List of positions or amino acids for each feature.

        Notes
        -----
        * Length parameters (``tmd_len``, ``jmd_n_len``, ``jmd_c_len``) must match with ids in ``features``.
        * Length of sequence (``tmd_seq``, ``jmd_n_seq``, ``jmd_c_seq``) must match with ids in ``features``.

        Examples
        --------
        .. include:: examples/sf_get_feature_positions.rst
        """
        # Check input
        features = ut.check_features(features=features)
        ut.check_number_val(name="start", val=start, just_int=True, accept_none=False)
        args_len, args_seq = check_parts_len(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len,
                                             tmd_seq=tmd_seq, jmd_n_seq=jmd_n_seq, jmd_c_seq=jmd_c_seq)
        check_match_features_seq_parts(features=features, **args_seq, **args_len)
        # Get feature position
        if args_seq["tmd_seq"] is not None:
            list_aa = get_amino_acids_(features=features, **args_seq)
            return list_aa
        else:
            list_pos = get_positions_(features=features, start=start, **args_len)
            return list_pos

    @staticmethod
    def get_df_pos(df_feat: pd.DataFrame = None,
                   col_val: str = "mean_dif",
                   col_cat: str = "category",
                   start: int = 1,
                   tmd_len: int = 20,
                   jmd_n_len: int = 10,
                   jmd_c_len: int = 10,
                   list_parts: Optional[Union[str, List[str]]] = None,
                   normalize : bool = False,
                   ) -> pd.DataFrame:
        """
        Create DataFrame of aggregated (mean or sum) feature values per residue position and scale.

        Parameters
        ----------
        df_feat : pd.DataFrame, shape (n_features, n_feature_info)
            Feature DataFrame with a unique identifier, scale information, statistics, and positions for each feature.
        col_val : {'abs_auc', 'abs_mean_dif', 'mean_dif', 'std_test', 'std_ref'}, default='mean_dif'
            Column name in ``df_feat`` containing numerical values to ``average``. If feature importance and impact
            are provided as {'feat_importance', 'feat_impact'} columns, their ``sum`` of values is computed.
        col_cat : {'category', 'subcategory', 'scale_name'}, default='category'
            Column name in ``df_feat`` for categorizing the numerical values during aggregation.
        start : int, default=1
            Position label of first residue position (starting at N-terminus).
        tmd_len : int, default=20
            Length of TMD (>0).
        jmd_n_len : int, default=10
            Length of JMD-N (>=0).
        jmd_c_len : int, default=10
            Length of JMD-C (>=0).
        list_parts: str or list of str, optional
            Specific sequence parts to consider for numerical value aggregation.
        normalize : bool, default=False
            If ``True``, normalizes aggregated numerical values to a total of 100%.

        Returns
        -------
        df_pos : pd.DataFrame, shape (n_categories, n_positions)
            DataFrame with aggregated numerical values per position.

        Notes
        -----
        * Length parameters (``tmd_len``, ``jmd_n_len``, ``jmd_c_len``) must match with feature ids in ``df_feat``.

        Examples
        --------
        .. include:: examples/sf_get_df_pos.rst
        """
        # Check input
        list_parts = ut.check_list_parts(list_parts=list_parts, return_default=False, accept_none=True)
        # Do not check for list_parts since df_pos can be obtained for any part
        df_feat = ut.check_df_feat(df_feat=df_feat)
        check_col_cat(col_cat=col_cat)
        check_col_val(col_val=col_val)
        ut.check_number_val(name="start", val=start, just_int=True, accept_none=False)
        args_len, _ = check_parts_len(tmd_len=tmd_len, jmd_n_len=jmd_n_len, jmd_c_len=jmd_c_len)
        ut.check_bool(name="normalize", val=normalize)
        check_match_features_seq_parts(features=df_feat[ut.COL_FEATURE], **args_len)
        # Get df pos
        stop = start + jmd_n_len + tmd_len + jmd_c_len - 1
        value_type = ut.DICT_VALUE_TYPE[col_val]
        df_pos = get_df_pos_(df_feat=df_feat, col_cat=col_cat, col_val=col_val, value_type=value_type,
                             start=start, stop=stop)
        if normalize:
            df_pos = df_pos / abs(df_pos).sum().sum() * 100
        if list_parts is not None:
            df_pos = get_df_pos_parts_(df_pos=df_pos, value_type=value_type,
                                       start=start, **args_len, list_parts=list_parts)
        return df_pos



"""This is a script for the backend of the TreeModel.predict_proba() method."""
import numpy as np

# II Main methods
def monte_carlo_predict_proba(X=None, list_models=None, is_selected=None):
    """Obtain Monta Carlo estimate of prediction probability of positive class"""
    # Initialize array to store predictions from all rounds
    all_predictions = np.zeros((len(list_models), X.shape[0]))
    # Iterate through each round and each model to make predictions
    for round_idx, models in enumerate(list_models):
        round_predictions = np.zeros((len(models), X.shape[0]))
        for model_idx, model in enumerate(models):
            # Select features for the current round and model
            selected_features = is_selected[round_idx, :]
            X_selected = X[:, selected_features]
            # Make predictions with the selected model and features
            # We are interested only in the second column (probability of positive class)
            round_predictions[model_idx] = model.predict_proba(X_selected)[:, 1]
        # Aggregate predictions for this round
        all_predictions[round_idx] = np.mean(round_predictions, axis=0)
    # Aggregate predictions across all rounds and compute standard deviation
    pred = np.mean(all_predictions, axis=0)
    pred_std = np.std(all_predictions, axis=0)
    return pred, pred_std



"""
This is a script for the backend of the dPULearn.eval() method.
"""
import numpy as np
import pandas as pd

import aaanalysis.utils as ut


# I Helper functions
# Test homogeneity in identified negatives
def _comp_std(X):
    """Calculate the Coefficient of Variation (CV) of the given X."""
    std_devs = np.std(X, axis=0)
    avg_std = np.mean(std_devs)
    return avg_std

def _comp_iqr(X):
    """
    Calculate the average interquartile range (IQR) of the given X.
    """
    q1 = np.percentile(X, 25, axis=0)
    q2 = np.percentile(X, 75, axis=0)
    # Calculate the IQR for each feature
    iqr_values = q2 - q1
    # Calculate the average IQR across all features
    avg_iqr = np.mean(iqr_values)
    return avg_iqr


# Test similarity between identified negatives and other classes
def _comp_auc(X=None, labels=None, label_test=0, label_ref=1):
    """Calculate the adjusted Area Under the Curve (AUC) of the given data."""
    X = X.copy()
    # Create a mask for the test and reference labels
    mask = np.asarray([l in [label_test, label_ref] for l in labels])
    # Compute AUC
    auc_abs_vals = abs(ut.auc_adjusted_(X=X[mask], labels=labels[mask], label_test=label_test))
    # Compute the average AUC
    avg_auc_abs = np.mean(auc_abs_vals)
    return avg_auc_abs


def _comp_kld(X, labels, label_test=0, label_ref=1):
    """Calculate the average Kullback-Leibler Divergence (KLD) for each feature."""
    kld_values = ut.kullback_leibler_divergence_(X=X, labels=labels, label_test=label_test, label_ref=label_ref)
    # Compute the average KLD
    avg_kld = np.mean(kld_values)
    return avg_kld


# II Main functions
def _eval_homogeneity(X=None, labels=None, label_test=0):
    """Compute two homogeneity measures of coefficient if variation (CV) and entropy"""
    mask = np.asarray([l == label_test for l in labels])
    avg_std = _comp_std(X[mask])
    avg_iqr = _comp_iqr(X[mask])
    return avg_std, avg_iqr


def _eval_distribution_alignment(X=None, labels=None, label_test=0, label_ref=1, comp_kld=True):
    """Compute the similarity between identified negatives and the other dataset classes (positives, unlabeled)"""
    # Perform tests
    avg_auc_abs = _comp_auc(X=X, labels=labels, label_test=label_test, label_ref=label_ref)
    avg_kld = _comp_kld(X=X, labels=labels, label_test=label_test, label_ref=label_ref) if comp_kld else None
    return avg_auc_abs, avg_kld


def _eval_distribution_alignment_X_neg(X=None, labels=None, X_neg=None, comp_kld=True):
    """Compute the similarity between identified negatives and ground-truth negatives"""
    label_test = 0
    label_ref = 1  # temporary label for ground-truth negatives for comparison
    mask_test = np.asarray([l == label_test for l in labels])
    # Create a combined dataset and labels for identified and ground-truth negatives
    X_test = X[mask_test]
    X_combined = np.vstack([X_test, X_neg])
    labels_combined = np.array([label_test] * len(X_test) + [label_ref] * len(X_neg))
    # Perform tests
    avg_auc = _comp_auc(X=X_combined, labels=labels_combined, label_test=label_test, label_ref=label_ref)
    avg_kld = _comp_kld(X=X_combined, labels=labels_combined, label_test=label_test, label_ref=label_ref) if comp_kld else None
    return avg_auc, avg_kld


@ut.catch_runtime_warnings()
def eval_identified_negatives(X=None, list_labels=None, names_datasets=None, X_neg=None, comp_kld=True):
    """Evaluate set of identified negatives for homogeneity and alignment with other datasets"""
    unl_in = False
    list_evals = []
    for labels in list_labels:
        # Evaluate homogeneity
        n_rel_neg = sum(labels == 0)
        avg_std, avg_iqr = _eval_homogeneity(X=X, labels=labels)
        # Evaluate distribution alignment with positives
        args = dict(X=X, labels=labels, comp_kld=comp_kld)
        avg_auc_abs, avg_kld = _eval_distribution_alignment(**args, label_test=0, label_ref=1)
        list_eval = [n_rel_neg, avg_std, avg_iqr, avg_auc_abs, avg_kld]
        # Evaluate distribution alignment with unlabeled
        if 2 in labels:
            avg_auc_abs, avg_kld = _eval_distribution_alignment(**args, label_test=0, label_ref=2)
            list_eval += [avg_auc_abs, avg_kld]
            unl_in = True
        # Evaluate distribution alignment with ground-truth negatives, if provided
        if X_neg is not None:
            avg_auc_abs, avg_kld = _eval_distribution_alignment_X_neg(**args, X_neg=X_neg)
            list_eval += [avg_auc_abs, avg_kld]
        # Remove kld if disabled
        list_eval = [x for x in list_eval if x is not None]
        list_evals.append(list_eval)
    # Define column names based on the evaluations performed
    cols_eval = ut.COLS_EVAL_DPULEARN[0:4]
    if comp_kld:
        cols_eval.append(ut.COL_AVG_KLD_POS)
    if unl_in:
        cols_eval.append(ut.COL_AVG_ABS_AUC_UNL)
        if comp_kld:
            cols_eval.append(ut.COL_AVG_KLD_UNL)
    if X_neg is not None:
        cols_eval.append(ut.COL_AVG_ABS_AUC_NEG)
        if comp_kld:
            cols_eval.append(ut.COL_AVG_KLD_NEG)
    # Create the DataFrame
    df_eval = pd.DataFrame(list_evals, columns=cols_eval).round(4)
    df_eval = ut.add_names_to_df_eval(df_eval=df_eval, names=names_datasets)
    return df_eval


"""
This is a script for the backend of the CPPPlot.eval() method.
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

import aaanalysis.utils as ut

BAR_WIDTH = 0.8
COLOR_BASE = "tab:gray"

# Helper functions
def _plot_n_features(ax=None, df_eval=None, dict_color=None):
    """Plots stacked bar charts for multiple sets on a single Axes."""
    names = df_eval[ut.COL_NAME].to_list()
    list_n_feat_sets = [x[1] for x in df_eval[ut.COL_N_FEAT]]
    n_sets = len(list_n_feat_sets)
    categories = list(dict_color.keys())
    # Set the width of the bars
    for i, list_n_feat in enumerate(list_n_feat_sets):
        bottom_pos = 0  # Start at 0 for each new set
        for j, category in enumerate(categories):
            counts = list_n_feat[j]
            # Stack the bars for each category
            ax.barh(i, counts, left=bottom_pos, height=BAR_WIDTH*0.66, color=dict_color[category],
                    label=category if i == 0 else "")
            bottom_pos += counts  # Increase the left offset by the count just plotted
    ax.set_yticks(np.arange(n_sets))
    ax.set_yticklabels(names)
    ax.set_xlabel(ut.COL_N_FEAT)
    sns.despine()



def _plot_range_abs_auc(ax=None, df_eval=None):
    """Boxplot for abs AUC ranges"""
    names = df_eval[ut.COL_NAME].to_list()
    range_abs_auc = df_eval[ut.COL_RANGE_ABS_AUC].tolist()
    # Boxplot
    boxprops = dict(facecolor=COLOR_BASE, color='black')
    arg = dict(color='black')
    ax.boxplot(range_abs_auc, positions=np.arange(len(names)), vert=False,
               patch_artist=True, boxprops=boxprops,
               whiskerprops=arg, capprops=arg, medianprops=arg, showfliers=False)
    # Scatter plot for min and max values
    min_total = None
    max_total = None
    for i, quin in enumerate(range_abs_auc):
        # quin is a list with the 5-number summary: [min, 25%, 50%, 75%, max]
        min_val = quin[0]
        max_val = quin[-1]
        ax.scatter(min_val, i, color=COLOR_BASE, edgecolor="black", zorder=3, s=40)
        ax.scatter(max_val, i, color=COLOR_BASE, edgecolor="black", zorder=3, s=40)
        if min_total is None or min_val < min_total:
            min_total = min_val
        if max_total is None or max_val > max_total:
            max_total = max_val
    ax.set_yticks(np.arange(len(names)))
    ax.set_yticklabels(names)
    x_adjust = (max_total - min_total)*0.1
    ax.set_xlim(min_total-x_adjust, max_total+x_adjust)
    ax.set_xticks([round(min_total, 2), round(max_total, 2)])
    ax.set_xticklabels([f'{min_total:.1f}', f'{max_total:.1f}'])
    ax.set_xlabel('abs AUC\n')
    sns.despine()


def _plot_avg_mean_dif(ax=None, df_eval=None):
    """Hued bar plot for average mean_dif"""
    dict_color = ut.plot_get_cdict_()
    color_pos = dict_color["FEAT_POS"]
    color_neg = dict_color["FEAT_NEG"]
    # Extract the tuples of mean differences
    list_mean_dif = df_eval[ut.COL_AVG_MEAN_DIF].tolist()
    names = df_eval[ut.COL_NAME].to_list()
    # Set the width of the bars
    bar_height = BAR_WIDTH/2
    # Plotting
    fs = ut.plot_gco()
    args_in = dict(va='center', ha='right', color="white", fontsize=fs-1)
    args_out = dict(va='center', ha='left', color="black", fontsize=fs-1)
    max_abs_dif = max([max([abs(x[0]), abs(x[1])]) for x in list_mean_dif])
    for i, (pos, neg) in enumerate(list_mean_dif):
        i_pos = i + bar_height / 2
        ax.barh(i_pos, pos, height=bar_height, color=color_pos)
        if pos > max_abs_dif/2:
            ax.text(pos, i_pos, f'{pos:.2f}', **args_in)
        else:
            ax.text(pos, i_pos, f'{pos:.2f}', **args_out)
        i_neg = i - bar_height / 2
        ax.barh(i_neg, abs(neg), height=bar_height, color=color_neg)
        if abs(neg) > max_abs_dif/2:
            ax.text(abs(neg), i_neg, f'{neg:.2f}', **args_in)
        else:
            ax.text(abs(neg), i_neg, f'{neg:.2f}', **args_out)
    # Setting the labels and title
    ax.set_yticks(np.arange(len(names)))
    ax.set_yticklabels(names)
    ax.set_xlabel('mean dif\n(avg pos/neg)')


def _plot_n_clusters(ax=None, df_eval=None):
    """Plot number of clusters"""
    list_n_clusters = df_eval[ut.COL_N_CLUST].tolist()
    names = df_eval[ut.COL_NAME].to_list()
    # Plotting
    n_max = max(list_n_clusters)
    for i, n_cluster in enumerate(list_n_clusters):
        ax.barh(i, n_cluster, height=BAR_WIDTH*0.66, color=COLOR_BASE)
        if n_cluster > n_max/2:
            ax.text(n_cluster, i, n_cluster, va='center', ha='right', color="white")
        else:
            ax.text(n_cluster, i, n_cluster, va='center', ha='left', color="black")
    # Setting the labels and title
    ax.set_yticks(np.arange(len(names)))
    ax.set_yticklabels(names)
    ax.set_xlabel('n_clusters')

def _plot_feat_per_cluster(ax=None, df_eval=None):
    """Plot average number of features per cluster"""
    list_avg_n_feat = df_eval[ut.COL_AVG_N_FEAT_PER_CLUST].tolist()
    list_std_n_feat = df_eval[ut.COL_STD_N_FEAT_PER_CLUST].tolist()
    names = df_eval[ut.COL_NAME].to_list()
    # Plotting
    for i, (avg_n_feat, std_n_feat) in enumerate(zip(list_avg_n_feat, list_std_n_feat)):
        ax.barh(i, avg_n_feat, height=BAR_WIDTH*0.66, color=COLOR_BASE)
        ax.errorbar(avg_n_feat, i, xerr=std_n_feat, ecolor="black", elinewidth=1,
                    capsize=5, capthick=1)
    # Setting the labels and title
    x_min, x_max = ax.get_xlim()
    ax.set_xlim(0, x_max)
    ax.set_yticks(np.arange(len(names)))
    ax.set_yticklabels(names)
    ax.set_xlabel('n feat/clust\n(avg ± std)')

# II Main functions
def plot_eval(df_eval=None, figsize=(8, 5), dict_xlims=None, dict_color=None, legend=True, legend_y=-0.3, list_cat=None):
    """Plot evaluation of CPP feature sets"""
    # Reverse order to have first dataset on top
    _df_eval = df_eval.iloc[::-1].reset_index(drop=True)
    fig, axes = plt.subplots(1, 5, sharey=True, figsize=figsize)
    # Plotting individual components
    _plot_n_features(ax=axes[0], df_eval=_df_eval, dict_color=dict_color)
    _plot_range_abs_auc(ax=axes[1], df_eval=_df_eval)
    _plot_avg_mean_dif(ax=axes[2], df_eval=_df_eval)
    _plot_n_clusters(ax=axes[3], df_eval=_df_eval)
    _plot_feat_per_cluster(ax=axes[4], df_eval=_df_eval)
    # Set xlims
    if dict_xlims is not None:
        for i in dict_xlims:
            axes[i].set_xlim(dict_xlims[i])
    # Adding central titles
    x_min = axes[1].get_xlim()[0]
    axes[1].set_title('Discriminative Power', ha='left', va='center', x=x_min, weight="bold")
    axes[3].set_title('Redundancy', ha='left', va='center', weight="bold")
    # Set legend under plot
    n_feat_cat = np.array([x[1] for x in df_eval[ut.COL_N_FEAT]]).sum(axis=0)
    # Filter non-occurring categories
    _list_cat = [cat for cat, n in zip(dict_color, n_feat_cat) if n != 0]
    if list_cat is None:
        list_cat = _list_cat
    else:
        list_cat = [x for x in list_cat if x in _list_cat]
    plt.tight_layout()
    if legend:
        fs = ut.plot_gco()
        ut.plot_legend_(ax=axes[0], dict_color=dict_color, list_cat=list_cat, loc="upper left",
                        n_cols=np.ceil(len(list_cat) / 2), y=legend_y,
                        title="Scale category", fontsize=fs-1,
                        labelspacing=0.1, columnspacing=0.4, handletextpad=0.2, handlelength=2)
        plt.subplots_adjust(wspace=0.25, hspace=0, bottom=0.35)
    else:
        plt.subplots_adjust(wspace=0.25, hspace=0)
    return fig, axes
